

/*  Agent0 Interpreter
 *   File #4 : int.com
 *   Commitment Rules
 */


/*   Commitment Rules are stored as :
 *
 *   commit(Msgcond,Mntlcond,Agent,Action)
 *
 *   where Agent is the agent for which the commitment rule is being
 *   made (since more than one agent can be in the system at one time)
 *   Msgcond : 
 *          [performative,keyword(value)] .....
 *   Mntlcond is of the form :
 *          [[Time1,Fact1],[Time2,Fact2],[Time3,Fact3]]
 *    which means that Agent needs to believe  (Fact1 @ Time1 ^ 
 *        Fact2 @ Time2 ^ Fact3 @ Time3) to commit to doing Action 
 *   Agent2 is the agent being committed to
 *   Action is of the form :
 *
 *          do(T,p-action).                 
 *          refrain(p-action).     *Equals do_not & removes UNREQUEST.
 *          if-then.
 *          believe([T,Fact]).     *NEW*
 *          Performative Messages  *INFORM/REQUEST/UNREQUEST delete, use KQML.
 *
 *
 *  Commitments are stored as :
 *
 *   cmt(Agent,Agent2,Action)
 *
 * Agents also believe all of their commitments -
 *
 *   cmt(Agent,Agent2,Action) -> b(Agent,cmt(Agent,Agent2,Action),[Time,t])
 * 
 * where Time is the time that the commitment was made.
 */

:- dynamic cmt/3, can/2, refrain/2, refrain/3.

refrain(noone,nothing).

/*
 * Deal with nested Actions in IF statements
 *
 */

can(Agent,if(_Mntl,Action)) :- can(Agent,Action).

/*
 *
 * This does the donkey work of check a single message against a single
 * commit rule. 
 *
 *
 */

process_one_message(Message,[commit(Msgcond,Mntcond,Sender1,Action)|Rest],X) :-
	       who_am_i(Me),    	       
               clock(Now),
	       Action =.. [Name|_],
	       dnl,dlog(X),dwrite('   '),dlog([Msgcond,Mntcond]),
	       dwrite(' => '),dlog(Name),
               ((message_match(Message,Msgcond),MTV=t);MTV=f),
               process_mental_cond(Me,Mntcond,Holds),
	       ((
                 (
                  (member(sender(Sender1),Message),Sender=Sender1);
                  (Sender1=nil,Sender=Me)
                 ),STV=t
                 )
                 ;STV=f),
               ((can(Me,Action),CTV=t);CTV=f),
	       ((not(refrain(Me,Action)),RTV=t);RTV=f),
	       ((MTV=t,Holds=t,STV=t,CTV=t,RTV=t,AllHolds=t);AllHolds=f),
               make_commitment(Me,Sender,Now,Action,AllHolds),
               X1 is X + 1,
	       process_one_message(Message,Rest,X1).


process_one_message(_Message,[],_).

/*
 * Check a list of Messages against the commit rules.
 *
 */

/*
 * 
 *  As above, but passes the complete ruleset. Empty set to deal with 
 *  
 */ 
process_messages([Message|Messages]) :- bagof(commit(Msgcond,Mntcond,Sender,Action),	
    		                        commit(Msgcond,Mntcond,Sender,Action),
			                Ruleset),
					linedash,
					dnl,dlognl(Message), 
	                                process_one_message(Message,Ruleset,1),
			 	        process_messages(Messages).

process_messages([]) :-		        bagof(commit(Msgcond,Mntcond,Sender,Action),
		                        commit(Msgcond,Mntcond,Sender,Action),
			                Ruleset),
                                        linedash,
				        dnl,dlognl([nil]),
					process_one_message([nil],Ruleset,1),
					dnl,dnl.


/*
 * 
 * Checks that the message condition component of a commit rule matches
 * a message, and thus may fire it.
 * 
 */ 
message_match(_Message, []).
message_match(Message, [Cond|Msgcond]) :- member(Cond,Message), 
					  message_match(Message,Msgcond).
/*
 *
 * Check the Mental Condition part of a Commit rule holds, this works
 * better than old version as it completely instantiates Variables etc,
 * and can call standard prolog.
 *
 */

process_mental_cond(_Agent,[],t).

process_mental_cond(Agent,[First|Rest],TVO) :- 
                                          process_mental_cond(Agent,First,TVF),
                                          process_mental_cond(Agent,Rest,TVR),
					  ((TVF=t,TVR=t,TVO=t);TVO=f).

process_mental_cond(Agent,not(Mntlcond),TV) :-
                            process_mental_cond(Agent,Mntlcond,OTV),
                            invert_truth(OTV,TV),!.

process_mental_cond(Agent,b([Time,Belief]),TV) :-
                       believes(Agent,[Time,Belief],TV).


process_mental_cond(_Agent,Predicate,TV) :-
				(call(Predicate),TV=t);TV=f.



/*
 *
 * What to do when a Commit Rule holds
 *
 */

make_commitment(_Agent,_Agent2,_Time,_Action,f).

make_commitment(_Agent,_Agent2,_Time,_Action,u).

make_commitment(Agent,Agent2,Time,Action,t) :-
              assertz(cmt(Agent,Agent2,Action)),
	      dnl,dnl,dwrite('       => '),dlog(Action),dnl,
              add_belief(Agent,[Time,cmt(Agent,Agent2,Action)]).



/*
 * Execute - Execute all the commitments that are due.
 *
 */


execute                 :- 
                           who_am_i(Me),                           
			   cmt(Me,Agent2,Action),
                           check_one_commitment(Me,Agent2,Action,_Flag),
                           fail;true. 
                           

check_one_commitment(Agent,Agent2,if(Mntlcond,Action),t) :-
                      process_mental_cond(Agent,Mntlcond,f),
                      safe_retract(cmt(Agent,Agent2,Action)),!.

check_one_commitment(Agent,Agent2,if(Mntlcond,Action),t) :-
                      process_mental_cond(Agent,Mntlcond,u),
                      safe_retract(cmt(Agent,Agent2,Action)),!.

check_one_commitment(Agent,Agent2,if(Mntlcond,Action),Flag) :-
                      process_mental_cond(Agent,Mntlcond,t),
                      check_one_commitment(Agent,Agent2,Action,Flag2),
                      Flag2 = f,Flag = f,!; 
                      safe_retract(cmt(Agent,Agent2,if(Mntlcond,Action))),
                      Flag = t.
                     
check_one_commitment(Agent,Agent2,do(Time,PrivateAction),Flag) :-
                      what_time_is_it(Now),
                      less(Now,Time),Flag = f,!;
                      call(PrivateAction),
                      safe_retract(cmt(Agent,Agent2,do(Time,PrivateAction))),
                      Flag = t,!.


check_one_commitment(Agent,Agent2,refrain(Time,Agent3,Action),Flag) :-
                      what_time_is_it(Now),
                      less(Now,Time),Flag = f,!;
                      assertz(refrain(Agent,Action)),
                      safe_retract(cmt(Agent,Agent2,refrain(Time,Agent3,Action))),
                      Flag = t,!.


check_one_commitment(Agent,Agent2,believe([Time,Fact]),Flag) :-
                      add_belief(Agent,[Time,Fact]),
                      add_belief(Agent,[Time,b(Agent2,Fact)]),
                      safe_retract(cmt(Agent,Agent2,believe([Time,Fact]))),
                      Flag = t,!.

check_one_commitment(Agent,Agent2,kqml(Time,[Perf|Message]),Flag) :-
                      what_time_is_it(Now),
                      less(Now,Time),Flag = f,!;
                      send_message([Perf|Message]),
                      safe_retract(cmt(Agent,Agent2,kqml(Time,[Perf|Message]))),
                      Flag = t,!.



safe_retract(A) :- not(A);
                   retract(A).


