%-----------------------------------------------------------------------------
% kqmlparse.pl  - Winton Davies, (c) University of Aberdeen, 1994.
%-----------------------------------------------------------------------------
% v1.1
%-----------------------------------------------------------------------------
% pltokqml(+Prolog, -KQML)
% kqmltopl(+KQML, -Prolog)
% kqml_prolog(?KQML,?Prolog)
%
% Converts a string from kapi interface into a Prologish interpretation of
% KQML performatives. Note that the incoming string is a Prolog 'atom'. 
%
% The Prolog version of KQML is in the form:
%  
%   [Performative,keyword(Value),...]
%
% Note that it handles multiply nested messages.
%
% Restrictions:
% 
% (1) Only handles contents of Prolog and KQML
% (2) Keyword values other than :content really restricted to words.
%     This could be changed I guess - but to what language ?
% (3) Uses unsupported Quintus Library to handle Prolog parsing.
% (4) complex prolog structures should be terminated by a fullstop
%     when being sent through KQML. Messages generated by this agent
%     will add the fullstop. Nasty, but from use of read(X).
%     Could be remedied through the use of a full Prolog DCG.
% (5) Note that now I have fixed the bug that refused to handle special chars
%     that parsing prolog could be weird (I doesn't appear to be from tests).
%
% (6) addrMap stores name;number form.... uses address to convert... 
%     inwards looks for sender;1 outwards looks for receiver receiver.
%-----------------------------------------------------------------------------
%
% Use non-standard Quintus library for quick hack to parse prolog.
%
:- ensure_loaded(library(fromonto)).
:- dynamic addrMap/2.
%-----------------------------------------------------------------------------
%
% Basic BNF performatives described by KQML draft. 
%
% Strings,Quotes and Complexes are non-Prolog, so we'll skip that....
%----------------------------------------------------------------------------- 
kqmltopl(KQML,Prolog) :- name(KQML,KQMLstring),
	                kqmltopl(Prolog,KQMLstring,[]).

kqmltopl([Perf|KList])-->"(",
                              perf(Perf), 
                              kywrd_vals(KList),
                              white_space,")",
                              {member(language(prolog),KList);
                               member(language(kqml),KList)}.

kqmltopl([error,message(langnotsupported)])-->"(",
                              perf(_Perf), 
                              kywrd_vals(_KList),
                              white_space,")".

perf(Perf) --> white_space, word(Perf).

kywrd_vals([KWV|Rest]) --> kywrd_val(KWV), kywrd_vals(Rest).
kywrd_vals([]) --> [].

kywrd_val(KWV) --> keyword(KW), value(V),{KWV =.. [KW,V]}.

keyword(KW) --> compulsory_white_space, ":", word(KW).
value(V) --> compulsory_white_space, expression(V).

%-----------------------------------------------------------------------------
% Basic components.
%-----------------------------------------------------------------------------

expression(E) --> word(E).
expression(E) --> kqml(E).
expression(E) --> prolog(E).

%
% All of these not much use to Prolog - so, we'll just glaze over them...
% A full implementation should really use KIF, and translate to and
% from Prolog.... But I'm not superman.
%
%expression(E) --> quotation(E).
%expression(E) --> string(E).
%expression(E) --> complex_expression(E).

word(Word) --> character(C), characters(Chars), {name(Word, [C|Chars])}.
kqml(P) --> kqmltopl(P).

% Hush up the syntax errors generated if fails to parse the entire message.
prolog(P) --> asciis(List),{prolog_flag(syntax_errors,X, quiet),
                            (prolog_string(P,List) ;
                             prolog_flag(syntax_errors,_,X),fail),
                            prolog_flag(syntax_errors,_,X)}.


ascii(C) --> [C],{C > 0,C < 127}.
asciis([C|Chars]) --> ascii(C), asciis(Chars).
asciis([]) --> [].

character(C) --> alphabetic(C);numeric(C);special(C).
characters([C|Chars]) --> character(C), characters(Chars).
characters([]) --> [].

alphabetic(C) --> [C], {C > 96, C < 123}.	% Lower case.
alphabetic(C) --> [C], {C > 64, C < 91}.	% Upper case.
numeric(C) --> [C], {C > 47, C < 58}.	% Digits.
special(C) --> [C], {member([C],["<",">","=","+","-","*","/","&","^","~","_","@","$","%",":",".", "!","?",";"])}.

compulsory_white_space -->  " " , white_space.   % space
%compulsory_white_space --> "	", white_space. % tab
white_space --> " ", white_space. 
white_space --> "	", white_space. 
white_space --> [].

member(X,[X|_]).
member(X,[_|Y]) :- member(X,Y).

prolog_string(Prolog,String) :- var(Prolog), nonvar(String),
                                read(Prolog) from_chars String.
prolog_string(Prolog,String) :- nonvar(Prolog), var(String),
                                klog(Prolog) onto_chars String.
	/*			append(String1,".",String). */



klog(Term) :- portray_clause(Term) onto_chars String,
              koutput_wo_nl(String).

koutput_wo_nl([]) :- !.
koutput_wo_nl([10|Rest]) :- koutput_wo_nl(Rest), !.
koutput_wo_nl([H|Rest]) :- put(H), koutput_wo_nl(Rest),!.


%-----------------------------------------------------------------------------
% pltokqml(+Prolog,-kqml).
%-----------------------------------------------------------------------------
pltokqml(PrologKQML,KQML) :-    pltokqml(PrologKQML,KQML,PrologKQML).

pltokqml([],'',_PrologKQML).

pltokqml([H|T],KQML,PrologKQML) :- atom(H), 
	                        name(H,Atom),
	                        pltokqml(T,KQML2,PrologKQML),
				name(KQML2,KQMLSTR2), 
                                append(KQMLSTR2,")",Tmpstr),
				append(Atom,Tmpstr,Tmpstr2),
				append("(",Tmpstr2,KQMLSTR),
				name(KQML,KQMLSTR).
% word  %				
pltokqml([H|T],KQML,PrologKQML) :- H =.. [Keyword,Value],
                                name(Keyword,KeyStr),
			        atom(Value),
				name(Value,ValStr),
				append(" ",ValStr,ValStr2),
				append(KeyStr,ValStr2,KeyValStr),
				append(" :",KeyValStr,KQMLSTR2),
				pltokqml(T,KQML3,PrologKQML),
				name(KQML3,KQMLSTR3), 
				append(KQMLSTR2,KQMLSTR3,KQMLSTR),
				name(KQML,KQMLSTR).
	
pltokqml([H|T],KQML,PrologKQML) :- H =.. [Keyword,Value],
                                name(Keyword,KeyStr),
				member(language(kqml),PrologKQML),
				pltokqml(Value,ValList),
				name(ValList,ValStr),
				append(" ",ValStr,ValStr2),
				append(KeyStr,ValStr2,KeyValStr),
				append(" :",KeyValStr,KQMLSTR2),
				pltokqml(T,KQML3,PrologKQML),
				name(KQML3,KQMLSTR3), 
				append(KQMLSTR2,KQMLSTR3,KQMLSTR),
				name(KQML,KQMLSTR).

pltokqml([H|T],KQML,PrologKQML) :- H =.. [Keyword,Value],
                                name(Keyword,KeyStr),
				member(language(prolog),PrologKQML),
				prolog_string(Value,ValStr),
				append(" ",ValStr,ValStr2),
				append(KeyStr,ValStr2,KeyValStr),
				append(" :",KeyValStr,KQMLSTR2),
				pltokqml(T,KQML3,PrologKQML),
				name(KQML3,KQMLSTR3), 
				append(KQMLSTR2,KQMLSTR3,KQMLSTR),
				name(KQML,KQMLSTR).
	


kqml_prolog(KQML,PROLOG) :- (nonvar(KQML),kqmltopl(KQML,PROLOG1),
	                           address(PROLOG1,PROLOG))
                                     ;
	                    (nonvar(PROLOG),address(PROLOG1,PROLOG), 
			           pltokqml(PROLOG1,KQML)).



address(With_Name_Num,With_Name) :- nonvar(With_Name_Num),
	                  append(H,[sender(S)|T],With_Name_Num),
			  name(S,NameNumStr),
			  append(NameStr,[59|_NumStr],NameNumStr),
			  name(Name,NameStr),
			  append(H,[sender(Name)|T],With_Name),
			  ((addrMap(Name,NameNum),
			    retract(addrMap(Name,NameNum)));true),
			  assert(addrMap(Name,S)).


address(With_Name_Num,With_Name) :- nonvar(With_Name),
	                  append(H,[receiver(S)|T],With_Name),
			  addrMap(S,NameNum),
			  append(H,[receiver(NameNum)|T],With_Name_Num).
			  
address(CatchAll,CatchAll).

/*
   Test Message

:- KQMLIn = '(reply-all :sender person;1 :receiver person :content (tell :content [A]. :language prolog) :language kqml)',
   kqml_prolog(KQMLIn, Prolog),    write(Prolog),nl,
   kqml_prolog(KQML,   Prolog),    write(KQML),nl,
   kqml_prolog(KQML,  Prolog2),    write(Prolog2),nl.
*/


