/*
 * kaSend.c -- KQML API send routines
 *
 * Copyright (c)  1993, 1994 Enterprise Integration Technologies Corporation
 * and Lockheed Missiles and Space Company, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Enterprise Integration Technologies Corporation and Lockheed Missiles
 * and Space Company may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission
 * of Enterprise Integration Technologies Corporation and Lockheed Missiles
 * and Space Company.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL ENTERPRISE INTEGRATION TECHNOLOGIES CORPORATION OR
 * LOCKHEED MISSILES AND SPACE COMPANY BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <kapi.h>
#include <kapi_int.h>


/*
 *----------------------------------------------------------------------
 *
 * sendOnTransport --
 *
 *	Sends given message to given parsed url destination
 *
 * Results:
 *	1  - if successful
 *      -1 - for failure
 *
 * Side effects:
 *      OpenConns hash table of open connections may be updated, and new
 *      connections may be established.
 *
 *----------------------------------------------------------------------
 */

int
#if defined(__STDC__) || defined(__cplusplus)
sendOnTransport(ParsedURL *purl, char *buf)
#else
sendOnTransport(purl, buf)
     ParsedURL *purl;
     char *buf;
#endif
{
  int (*SendFunc) _ANSI_ARGS_((ParsedURL *, char *));
  SendFunc = Transports[purl->transtype]->Send;
  return((*SendFunc)(purl, buf));
};


/*
 *----------------------------------------------------------------------
 *
 * KSendString --
 *
 *	Sends given message.  Parses message to find appropriate destination.
 *
 * Results:
 *	K_OK = success
 *      K_BAD_KQML = could not send bad kqml message
 *      K_NO_RECEIVER = no receiver was specified in kqml msg
 *      K_NO_TRANSPORT = could't find the transport for that agent name
 *
 * Side effects:
 *      OpenConns hash table of open connections may be updated, and new
 *      connections may be established.
 *
 *----------------------------------------------------------------------
 */

int
#if defined(__STDC__) || defined(__cplusplus)
KSendString(char* message)
#else
KSendString(message)
     char *message;        /* this is a KQML message to send */
#endif
{
    ParsedURL *purl = NULL;
    char buffer[K_MAX_MESSAGE_SIZE];
    char *connection_id = NULL;
    char *address = NULL;
    int namelen;
    char *strptr = NULL;
    HashEntry *entry = NULL;
    ParsedKQMLMess *pmessage = 
	(ParsedKQMLMess *) calloc(1, sizeof(ParsedKQMLMess));
   
    /* The following approach is used to keep non-ANSI compilers happy.
       gcc allows direct initializtion of the KTable as in
       KeydPair KTable[] = {
       {":sender", &(pmessage->sender)},
       {":receiver", &address},
       {(char *) NULL, (char **) NULL}
       }; */
    KeydPair KTable[3];
    char *key1 = ":sender";
    char *key2 = ":receiver";
    KTable[0].keyword = key1;
    KTable[0].value_address = &(pmessage->sender);
    KTable[1].keyword = key2;
    KTable[1].value_address = &address;
    KTable[2].keyword = (char *) NULL;
    KTable[2].value_address = (char **) NULL;
    
    bzero(buffer, K_MAX_MESSAGE_SIZE);

    if ((KParse(message, 
	       &(pmessage->performative), 
	       KTable, 
	       &(pmessage->other_pairs)) != K_OK) || 
	pmessage->performative == NULL) {
	/* Parse failure. Trying to send invalid KQML message */
#if (DEBUG > 0)
	fprintf(stderr,"Attempt to send unparseable KQML message. Send Aborted\n");
#endif
	return(K_BAD_KQML);
    }

    /* Check to see if a specific connection is identified in receiver field.
       If so, look it up and use it.  */
    if (address == NULL) {
	return K_NO_RECEIVER;
    }

    namelen = strcspn(address, ";");
    pmessage->receiver = (char *) malloc(namelen + 1);
    strncpy(pmessage->receiver, address, namelen);
    pmessage->receiver[namelen] = '\0';

    /* If a specific connection is identified in the sender field
       send message on that connection */

#if ADD_CONN_INFO
    if (namelen < strlen(address) &&
	(entry = FindHashEntry(&OpenConns, address + namelen + 1)) != NULL &&
	(purl = (ParsedURL *) GetHashValue(entry)) != NULL &&
	(purl->state == OPEN_CON) &&
	sendOnTransport(purl, ParsedMessToString(pmessage, buffer)) != -1) {
	DeletePMessage(pmessage);
	if (address) free(address);
	return(K_OK);
    }
#endif
  
    /* Must decide which transport to use to send string buf. */

    purl = getServerURL(pmessage->receiver);
    while (purl != NULL) {
	if (sendOnTransport(purl, ParsedMessToString(pmessage, buffer)) != -1) {
	    DeletePMessage(pmessage);
	    if (address) free(address);
	    return(K_OK);
	}
	purl = purl->alternate;
    }

    DeletePMessage(pmessage);
    if (address) free(address);
#if (DEBUG > 0)
    fprintf(stderr, "Send failure: Could not find a suitable transport.\n");
#endif
    return K_NO_TRANSPORT;
}


/*
 *----------------------------------------------------------------------
 *
 * getServerURL --
 *
 *	Finds a parsed url corresponding to given agent identifier.  Searches
 *      in the following sequence:
 *         1)  Look for open connection
 *         2)  See if agent id is in cache
 *         3)  See if agent id is a direct address (URL)
 *         4)  Query ANS for agent id
 *
 * Results:
 *	returns parsed url if successful
 *      returns NULL for failure
 *
 * Side effects:
 *      Results of ANS query are cached in hash table.
 *
 *----------------------------------------------------------------------
 */

ParsedURL *
#if defined(__STDC__) || defined(__cplusplus)
getServerURL(char *agentID)
#else
getServerURL(agentID)
char *agentID;
#endif
{
  int i, j;
  HashEntry *entry = NULL;
  ParsedURL *newURL = NULL;
  int numURLs, flg;

  if (agentID == NULL)
    return(NULL);

  /* Check to see if connection to agentID is open */
  if((entry = FindHashEntry(&OpenConns, agentID)) != NULL)
    return((ParsedURL *) GetHashValue(entry));

  /* Check to see if agentID is cached locally */
  if((entry = FindHashEntry(&Servers, agentID)) != NULL)
    return((ParsedURL *) GetHashValue(entry));

  /* Check to see if agentID is a direct address */
  if ((newURL = ParseURL(agentID)) != NULL)
    return(newURL);

  /* Query ANS (Agent Name Server) if available */
  if (gANSpurl == NULL) {
#if (DEBUG > 0)
    fprintf(stderr, "No ANS available. Abort send.\n");
#endif
    return(NULL);
  }
#if (DEBUG > 0)
  fprintf(stderr, "getServerURL: Querying ANS.\n");
#endif

  if((newURL = ANSLookup(agentID, gANSpurl)) == NULL) {
#if (DEBUG > 0)
    fprintf(stderr, "getServerURL: No transport for agent %s\n",
	    agentID);
#endif
    return(NULL);
  }

  entry = CreateHashEntry(&Servers, agentID, &flg);
  strcpy(newURL->name, agentID);
  SetHashValue(entry, newURL);

  return(newURL);
};


