"""blcommunication.py

Contains classes for communicating with the brokering layer (bl).

classes:
    BlConnection
"""
# standard modules
import os
import string

# omniORB
from omniORB import CORBA

# client side stubs
try:
    import piper.DL2BL
except ImportError:
    print 'Whoops. You need to remember to run omniidl on the idl file'
    print 'to generate the client stubs!'

#### constants
# the name of the environmental variable to get the IOR from
ENV_VAR_IOR = 'VSH_DL2BL_IOR'
# the location of the default IOR if the environmental variable isn't set
DEFAULT_IOR = os.path.join(os.environ['HOME'], 'piper_info', 'ior','dl2bl.ior')
print 'ior file: ',DEFAULT_IOR
class BlConnection:
    """Establish a connection with the brokering layer.
    """
        
    def __init__(self):
        """Initalize a connection with the bl corba server.
        
        Note: A reference to the server will be obtained by reading the 
        stringified IOR from a file.
        This location of this file is presumed to be defined in the 
        environmental variable VSH_DL2BL_IOR. It this environmental 
        variable is not set, the default location will be used 
        ('/tmp/__VSH_DL2BL.ior').
        
        Raises an IOError if the ior file can't be found.
        """
        # get the orb started
        orb = CORBA.ORB_init([], CORBA.ORB_ID)
        
        # get the location of the IOR file from the environmental variable
        try:
            ior_file_location = os.environ[ENV_VAR_IOR]
        # if the environmental variable is not found, get it from the 
        # default location.
        except KeyError:
            # print 'Environmental variable', ENV_VAR_IOR, 'not found.'
            # print 'Using default ior file location.'
            # default location of the ior file
            ior_file_location = DEFAULT_IOR
    
        # grab our string IOR from the file and convert it to an object
        # reference
        ior_file = open(ior_file_location, 'r')
        string_ior = ior_file.read()
        ior_file.close()
    
        # now convert the stringified ior into an object reference
        self.connection_ref = orb.string_to_object(string_ior)
        self.connection_ref._narrow(piper.DL2BL.Connection)
        
    def setup_connection(self, password, connect_type = 'plain'):
        """Establish a connection with the brokering layer.
        
        Arguments:
            * password - The password used to secure the connection.
            * connect_type - The connection type to initiate. Possible 
            types are:
                o 'unknown'
                o 'unconnected'
                o 'plain'
                o 'encrypted'
        Returns:
            * The id number assigned to the dl.
        """
        self.password = password
        
        # make the connection type string into the appropriate enum type
        if connect_type == 'plain':
            conn_to_request = piper.DL2BL.CONNPLAIN
        elif connect_type == 'encrypted':
            conn_to_request = piper.DL2BL.CONNENCRYPTED
        # don't know why anyone would want to request these types :-)
        # but I'll include 'em just to be complete.
        elif connect_type == 'unknown':
            conn_to_request = piper.DL2BL.CONNUNKNOWN
        elif connect_type == 'unconnected':
            conn_to_request = piper.DL2BL.CONNUNCONNECTED
        else:
            raise TypeError \
              ('The connect type ' + connect_type + ' is not defined')
        
        self.dl_id = 2
        dl_connected = 0
        # keep looping through until we get assigned a unique dl_id
        while dl_connected == 0:
            try:
                print "setting up a connection with a unique dl_id (%i) ..." % self.dl_id
                proc_struct = self.connection_ref.setup(self.dl_id, password,
                  conn_to_request)
                print "dl_id was accepted by BL"
                dl_connected = 1
            # if we get an IdInUse error we need to try another id
            except piper.DL2BL.IdInUse:
                self.dl_id = self.dl_id + 1
                
        # now get our references for uploaded representations and querying
        self.rep_ref = proc_struct.representation
        self.query_ref = proc_struct.obtain
        
        return self.dl_id
                
    def release_connection(self):
        """Release the connection with the brokering layer.
        
        Raises:
            IOError if the connection is not released.
            NameError if we have a problem and the id is not found.
        """
        try:
            was_success = self.connection_ref.release(self.dl_id,
                                                      self.password)
        except piper.DL2BL.IdNotFound:
            raise NameError("The id %s was not found in the bl." % self.dl_id)
        
        # if we were not successful, raise an error
        if was_success == 0:
            raise IOError('Attempt to release connection was not successful.')
        
    def check_connection(self):
        """Check the status of the connection with the brokering layer.
        
        Returns:
            * A string describing the status of the connection:
                o 'unknown'
                o 'unconnected'
                o 'plain'
                o 'encrypted'
                
        Raises:
            IOError - if the definition layer can't be queried.
        """
        try:
            con_status = self.connection_ref.check(self.dl_id)
        except piper.DL2BL.InfoNotAvailable:
            raise IOError \
              ('The definition layer cannot display the connection status')
        
        # convert the con_status into a string to pass back
        if con_status == piper.DL2BL.CONNPLAIN:
            con_to_return = 'plain'
        elif con_status == piper.DL2BL.CONNENCRYPTED:
            con_to_return = 'encrypted'
        elif con_status == piper.DL2BL.CONNUNKNOWN:
            con_to_return = 'unknown'
        elif con_status == piper.DL2BL.CONNUNCONNECTED:
            con_to_return = 'unconnected'
        
        return con_to_return
        
    def submit_representation(self, xml_representation):
        """Send an xml document to the brokering layer for processing.
        
        Arguments:
            * xml_representation - The full path to an xml file for submission
            to the bl.
            * password - The password used to authenticate the connection.
            
        Raises:
            IOError - If the brokering layer can't accept the document.
        """
        print 'dl_id = ',self.dl_id
        print 'password = ',self.password
        try:
            self.rep_ref.upload(xml_representation, self.dl_id, self.password)
        except piper.DL2BL.UriUndefined:
            raise IOError \
              ('One of the uris defined in the xml can not be found.')
            
    def update_representation(self, uri_to_update, xml_representation):
        """Update a submitted xml representation with a new representation.
        
        Arguments:
            * uri_to_update - The specific uri that should be replaced by 
            the new xml_representation. This should be as small as possible.
            So:
                o dl_id.0.0 - To update a whole document
                o dl_id.subnet_id.0 - To update a specific subnet is a document.
                o dl_id.subnet_id.node_id - To update a specific node.
                
        Raises:
            IOError - If the definition layer can't overwrite the uri, or if
            the update was unsuccessful for any reason.
        """
        # convert the uri to update into a struct to pass
        state = piper.DL2BL.stateS(piper.DL2BL.ERRORNONE, 1, 0, 0, 0)
        status = piper.DL2BL.statusS(state, 0)
        
        dl_id, subnet_id, node_id = string.split(uri_to_update, '.')
        uri_to_pass = piper.DL2BL.uriS(dl_id, subnet_id, node_id, status)
        try:
            was_success = self.rep_ref.update(xml_representation, uri_to_pass,
              self.dl_id, self.password)
        except piper.DL2BL.UriInUse:
            raise IOError \
              (uri_to_update + ' is being processed and can not be updated.')
              
        if was_success == 0:
            raise IOError('Update was not successful.')
              
    def download_representation(self, uri_to_download):
        """Download an xml representation from the bl.
        
        Arguments:
            * uri_to_download - The specific uri that should be downloaded. 
            This should be as specific as possible.
            So:
                o dl_id.0.0 - To download a whole document
                o dl_id.subnet_id.0 - To download a specific subnet in a 
                document.
                o dl_id.subnet_id.node_id - To download a specific node.
        
        Returns:
            * A huge string containing an xml representation.
            
        Raises:
            * IndexError - If the passed uri could not be found.
        """
        # convert the uri to download into a struct to pass
        state = piper.DL2BL.stateS(piper.DL2BL.ERRORNONE, 1, 0, 0, 0)
        status = piper.DL2BL.statusS(state, 0)
        
        dl_id, subnet_id, node_id = string.split(uri_to_download, '.')
        uri_to_pass = piper.DL2BL.uriS(dl_id, subnet_id, node_id, status)
        
        try:
            downloaded_xml = self.rep_ref.download(uri_to_pass, self.dl_id, 
              self.password)
        except piper.DL2BL.UriNotFound:
            raise IndexError\
              ('The passed uri ' + uri_to_download + ' was not found in the bl.')
        
        return downloaded_xml
        
    def get_program_ids(self):
        """Retrieve all of the program ids available in the bl.
        
        Returns:
            A list of available subnet and nodes. Subnets will be
            returned with an id like 0.subnet_id.0 while nodes will be 
            returned with ids like 0.0.node_id.
        """
        uri_ids = self.query_ref.programIds()
        # convert the uriS structs returned into strings to return
        uris_to_return = []
        for n in uri_ids:
            this_uri = str(n.instanceID) + '.' + str(n.subnetID) + '.' + str(n.nodeID) + ':::' + str(n.description)
            uris_to_return.append(this_uri)
            
        return uris_to_return
                
    def check_program(self, program_uri):
        """Check a particular program (node or subnet) to see if it is available.
        
        Arguments:
            program_uri - The uri representing the program.
        
        Returns:
            1 if the program is available
            0 if it is not.
            
        Raises:
            IOError - If the passed uri is not found in the bl.
        """
        # convert the uri to check into a struct to pass
        state = piper.DL2BL.stateS(piper.DL2BL.ERRORNONE, 1, 0, 0, 0)
        status = piper.DL2BL.statusS(state, 0)
        
        dl_id, subnet_id, node_id = string.split(program_uri, '.')
        uri_to_pass = piper.DL2BL.uriS(dl_id, subnet_id, node_id, status)
        
        try:
            is_available = self.query_ref.uriIsAvailable(uri_to_pass)
        except piper.DL2BL.UriNotFound:
            raise IOError('The uri ' + program_uri + ' was not found in the bl.')
            
        return is_available
        
    def query_status(self, uri_to_query):
        """Return the status of the passed uri.
        
        Arguments:
            * uri_to_query - The specific uri to query, in the uri string
            format of xx.xx.xx
            
        Returns:
            * A list of strings describing the status of the uri item
                o 'constructed' - A uri item that's standing by, 
                one that's fully realized.
                o 'blocking' - A uri item that's temporarily disabled. 
                Might be because it's currenly running or something.
                o 'temporary' - A uri item that exists temporarily, 
                mostly used to 'run' a subnet that's blocked by copying it and
                executing that copy.
                o 'isolated' - A uri item that cannot have any 
                connection\relation to another one.
            * A double value describing the process of the document towards
            completion (0 = hasn't started; 1 = complete).
            
        Raises:
            IndexError - If the uri isn't found in the brokering layer.
        """
        # convert the uri to query into a struct to pass
        state = piper.DL2BL.stateS(piper.DL2BL.ERRORNONE, 1, 0, 0, 0)
        status = piper.DL2BL.statusS(state, 0)
        
        dl_id, subnet_id, node_id = string.split(uri_to_query, '.')
        uri_to_pass = piper.DL2BL.uriS(dl_id, subnet_id, node_id, status)
            
        try:
            uri_status = self.query_ref.uriStatus(uri_to_pass, self.dl_id, 
              self.password)
        except piper.DL2BL.UriNotFound:
            raise IndexError \
              ('The uri ' + uri_to_query + ' was not found in the bl.')
              
        # build the returned uri_status into the proper info to return
        string_status = []
        if uri_status.state.constructed == 1:
            string_status.append('constructed')
        if uri_status.state.blocking == 1:
            string_status.append('blocking')
        if uri_status.state.temporary == 1:
            string_status.append('temporary')
        if uri_status.state.isolated == 1:
            string_status.append('isolated')
            
        return string_status, uri_status.progress
        
    def obtain_results(self, uri_for_results):
        """Return the results of processing a node.
        
        Arguments:
            uri_for_results - The uri to obtain the results from.
            
        Returns:
            A string with the info from the node.
            
        Raises:
            IOError - If the results from the node are not available.
        """
        # convert the uri to get results from into a struct to pass
        state = piper.DL2BL.stateS(piper.DL2BL.ERRORNONE, 1, 0, 0, 0)
        status = piper.DL2BL.statusS(state, 0)
        
        dl_id, subnet_id, node_id = string.split(uri_for_results, '.')
        uri_to_pass = piper.DL2BL.uriS(dl_id, subnet_id, node_id, status)
        
        try:
            results = self.query_ref.uriInfo(self.password, uri_to_pass)
        except piper.DL2BL.InfoNotAvailable:
            raise IOError('Results are not available for ' + uri_for_results)
            
        return results
              
        

        

        
        
