"""compositelocus.py

Classes describing the composite locus type.

Classes:
    CompositeLocus
    TempCompositeLocus
"""
# standard modules
import os
import string
import tempfile

# local modules
from containerlocus import ContainerLocus
from linklocus import SimpleLinkLocus
from locusid import LocusDb
from locus import Locus

# Overflow 
import vflow

# configuration
from piperconfig import *

### constants
TEMP_DIR = os.path.join(os.environ['HOME'], 'piper_info', 'temp')
BASE_XML_DIR = os.path.join(piper_dir, 'xml')
BASE_TB_DIR = overflow_toolbox_dir
OF_LIB = os.path.join(overflow_prefix, 'lib')

class CompositeLocus(ContainerLocus):
    def __init__(self, parent_locus, base_locus):
        ContainerLocus.__init__(self, parent_locus, base_locus)
        
        self._connected_loci = None
        
        self.of_links = []
        
    def initialize(self, of_net, of_node, name):
        """Initialize with the overflow network and node for this composite.

        'initialize()' should be called when a locus is already loaded in
        Overflow (ie. when we are loading a *.n file).
        """
        ContainerLocus.initialize(self, of_net, of_node, name)
        
    def create(self, type, name, do_init = 1):
        """Create a new locus of the specified container type.

        If 'do_init' is set at 1, the newly created locus will also be
        initialized by a call to 'initialize()'
        """
        valid_container_types = ['composite', 'iterator']
        if type not in valid_container_types:
            raise KeyError\
              ("The passed type %s is not a valid container type."% type)
        new_node, new_net = ContainerLocus.create(self, type, name, 0)
        
        if do_init:
            self.initialize(new_net, new_node, name)
                    
    # --- retrieval functions
    def get_children_loci(self):
        """Get all of the children loci held within the container locus.

        Returns:
        * A list with all of the children loci in this container. Each item
        to the list is the reference to the contained locus.
        """
        if self._children_loci is None:
            self._children_loci = {}
       
        return self._children_loci.values()
            
    # --- manipulation functions
        
    def create_locus(self, type, locus_name = None):
        """Create a new locus of the specified type.

        Arguments:
        * type - the type of the new locus, which must correspond to a
        type described in a loaded XML file.
        * locus_name - An optional argument specifying a unique name to
        be used for the locus.
        """
        if self._children_loci is None:
            self.get_children_loci()
        
        # if we are passed a locus name, use that
        if locus_name:
            self.base_locus.id_db.add_id(locus_name)
            new_locus_name = locus_name
        else:
            # get a new name for this locus
            new_locus_name = self.base_locus.id_db.get_new_id(type)
        
        if (type == 'composite') or (type == 'iterator'):
            new_t_locus = CompositeLocus(self, self.base_locus)
        elif (type == 'container'):
            new_t_locus = ContainerLocus(self, self.base_locus)
        else:
            new_t_locus = SimpleLinkLocus(self, self.base_locus)
            
        new_t_locus.create(type, new_locus_name)
        
       
        return new_t_locus
        
    def connect_loci(self, in_locus, in_connect_name, out_locus,
                     out_connect_name):
        """Connect two loci held in this composite locus.

        Arguments:
        * in_locus - The input locus of the link.
        * in_connect_name - The name of the connector in the input locus
        which will be involved in the connections.
        * out_locus - The output locus of the link.
        * out_connect_name - The name of the connector in the output locus
        which will be involved in the connection.
        """
        # be sure the connector names are uppercase
        in_connect_name = string.upper(in_connect_name)
        out_connect_name = string.upper(out_connect_name)
        
        # first update the input locus
        in_locus.add_link(in_locus, out_connect_name, in_connect_name, 
          'Input')
        
        # now the output locus
        out_locus.add_link(out_locus, in_connect_name, out_connect_name, 
          'Output')
          
        # add the link to the Overflow info
        # first find the input and out termini
        for input in in_locus.of_inputs:
            input_name = vflow.UITerminal_getName(input)
            if input_name == in_connect_name:
                in_term = input
                
        for output in out_locus.of_outputs:
            output_name = vflow.UITerminal_getName(output)
            if output_name == out_connect_name:
                out_term = output
        
        # now add the link
        new_link = vflow.UINetwork_newLink(self.of_net, out_term, in_term)
        self.of_links.append(new_link)
             
    def disconnect_loci(self, in_locus, in_connect_name, out_locus, 
      out_connect_name):
        """Disconnect two loci within the composite locus.

        Arguments:
        * in_locus - The input locus of the link to disconnect.
        * in_connect_name - The name of the connector in the input locus
        which should be disconnected.
        * out_locus - The output locus of the link to disconnect.
        * out_connect_name - The name of the connector in the output locus
        which should be disconnected.
        """
        # be sure the connector names are uppercase
        in_connect_name = string.upper(in_connect_name)
        out_connect_name = string.upper(out_connect_name)
        
        # first update the input locus
        in_locus.restore_link('Input', in_connect_name)
    
        # then the output locus
        out_locus.restore_link('Output', out_connect_name)
        
        # remove the link from the Overflow info
        # first find the input and out termini
        for input in in_locus.of_inputs:
            input_name = vflow.UITerminal_getName(input)
            if input_name == in_connect_name:
                in_term = input
                
        for output in out_locus.of_outputs:
            output_name = vflow.UITerminal_getName(output)
            if output_name == out_connect_name:
                out_term = output
                
        # now find the link to remove
        for link in self.of_links:
            in_link = vflow.UILink_getToTerminal(link)
            out_link = vflow.UILink_getFromTerminal(link)
            if (in_link == in_term) and (out_link == out_term):
                link_to_remove = link
        
        # now remove the link
        new_link = vflow.UINetwork_removeLink(self.of_net, link_to_remove)

    def add_connector(self, conn_type, old_name, new_name, link_type,
                      description, conn_locus):
        """Add a new connector to the network.

        This overrides the add_connector() function in linklocus.py to add
        support for adding netInputs and netOutputs to network.

        Arguments:
        * conn_type - The type of the new connector ('Input' or 'Output')
        * old_name - The name of the old node connector that we are associated
        * with.
        * new_name - The name of the new connector, which must be unique. 
        * link_type - A string specifying the type of information the connector
        will support.
        * description - A string providing information about the function of
        the connector.
        * conn_locus - The locus connected with this netInput or netOutput (ie.
        the locus whose connector feeds into the netInput or netOutput)

        Returns:
        * The new net terminal created for the connector.
        """
        # add the connector node information
        SimpleLinkLocus.add_connector(self, conn_type, new_name, link_type,
                                      description)

        # need to add the new connector in the overflow information
        
        # if we have a network, then we need to add this as a netInput
        # or netOutput
        if self.of_net:
            
            if conn_type == 'Input':
                is_input = 1
                term_type = 0 # UINetTerminal::INPUT
            else:
                is_input = 0
                term_type = 1 # UINetTerminal::OUTPUT

            link_info = vflow.new_ItemInfo()
            vflow.ItemInfo_set_name(link_info, old_name)
            vflow.ItemInfo_set_type(link_info, link_type)
            vflow.ItemInfo_set_description(link_info, description)

            new_terminal = vflow.new_UITerminal(link_info, conn_locus.of_node,
               is_input, 1, 1)

            new_net_term = vflow.UINetwork_newNetTerminal(self.of_net,
              new_terminal, term_type, new_name)

            return new_net_term

        else:
            raise NotImplementedError\
                  ("Trying to add a netInput or netOutput without a network.")

    def map_locus_outputs(self, locus_list):
        """Map all unconnected outputs in the passed loci as netOutputs.

        Arguments:
        locus_list - A list of all loci whose unconnected outputs should
        be made as NetOutputs of this network.

        Raises:
        * LookupError - If no outputs are mapped to net terminals.

        Returns:
        * A list of all net terminals created in mapping these outputs
        """
        # all net termini that are added
        added_net_terms = []

        # the current network terminals
        cur_net_terms = vflow.UINetwork_getTerminals(self.of_net)
        cur_term_names = []
        for net_term in cur_net_terms:
            term_name = vflow.UINetTerminal_getName(net_term)
            cur_term_names.append(term_name)
        
        for out_locus in locus_list:
            # retrieve of the unconnected outputs
            connected_outputs = out_locus.get_connected_links('Output')
            all_outputs = out_locus.get_connector_info('Output')

            for conn_out in connected_outputs:
                to_remove = None
                for test_out in all_outputs:
                    print "conn_out[0]", conn_out[0]
                    print "test_out[0]", test_out[0]
                    if conn_out[0] == test_out[0]:
                        to_remove = test_out
                # remove the connected output
                all_outputs.remove(to_remove)

            # filter out the outputs that are already mapped
            for name in cur_term_names:
                to_filter = None
                for output in all_outputs:
                    output_name = (out_locus.get_net_name() + '::' + output[0])
                    if name == output_name:
                        to_filter = output

                if to_filter:
                    all_outputs.remove(to_filter)
                    

            # now add the left over unconnected outputs as netOutputs
            for unconn_output in all_outputs:
                new_name = (out_locus.get_net_name() + '::' +
                            unconn_output[0])
                new_net_term = self.add_connector('Output', unconn_output[0],
                  new_name, unconn_output[1], unconn_output[2], out_locus)
                added_net_terms.append(new_net_term)

        # if we mapped no outputs we should raise an error
        print 'added_net_terms', added_net_terms
        print 'cur_net_terms', cur_net_terms
        if (len(added_net_terms) == 0 and len(cur_net_terms) == 0):
            raise LookupError("No unconnected outputs were found to be mapped")

        # return the net termini that were created
        return added_net_terms

    def unmap_net_terms(self, terms_to_unmap):
        """Remove all of the passed net terminals.

        Arguments:
        terms_to_unmap - A list of net_terminals that should be unmapped.
        """
        for term in terms_to_unmap:
            vflow.UINetwork_removeTerminal(self.of_net, term)
                              
    def _load(self, of_network, dl_net):
        """Load up a composite given information from an overflow network.

        This should be called when loading a network from an *.n file.
        """
        self.of_net = of_network
        
        # --- first recursively load all the nodes in this composite
        
        # find all of the nodes in the network and add appropriate loci
        all_nodes = vflow.UINetwork_getNodes(self.of_net)
        for node in all_nodes:
            # add the node to the dictionary of of nodes
            node_name = vflow.UINode_getName(node)
            self.all_of_nodes[node_name] = node
            
            # determine the type of the node so we know whether we need
            # to add a network or just a regular node
            node_type = vflow.UINode_getType(node)
            
            # if the node is a network
            if node_type in self.base_locus.all_of_networks.keys():
                # retrieve the network pointer
                this_network = self.base_locus.all_of_networks[node_type]
    
                # now load the network
                net_type = vflow.UINetwork_getType(this_network)
                net_name = vflow.UINetwork_getName(this_network)
                
                # register the name in the id database
                self.base_locus.id_db.add_id(net_name)
        
                new_net = CompositeLocus(self, self.base_locus)
                new_net.initialize(this_network, node, net_name)
                
                if net_type == 0:
                    string_net_type = 'composite'
                elif net_type == 1:
                    string_net_type = 'iterator'
                else:
                    raise NotImplementedError("Cannot load network type.")
                
                # add a dl_network to the current network and show it
                # in the user interface
                new_dl_net = dl_net.add_initialized_network(string_net_type,
                  new_net)
                
                new_net._load(this_network, new_dl_net)
                
            # if it is just a node
            else:
                node_name = vflow.UINode_getName(node)
                # register the name in the id database
                self.base_locus.id_db.add_id(node_name)
                
                new_locus = SimpleLinkLocus(self, self.base_locus)
                new_locus.initialize(node, node_name)
                
                # add a dl_locus to the current network and show it in
                # the user interface
                new_dl_locus = dl_net.add_initialized_locus(node_type,
                  new_locus)
                  
                #new_locus._load(node)
                
        # --- now load what we need for this network
        
        # connect loci that are linked
        self.of_links = vflow.UINetwork_getLinks(self.of_net)
        
        for link in self.of_links:
            from_term = vflow.UILink_getFromTerminal(link)
            from_node = vflow.UITerminal_getNode(from_term)
            from_conn_name = vflow.UITerminal_getName(from_term)
            from_locus = self._children_loci[from_node]
            
            to_term = vflow.UILink_getToTerminal(link)
            to_node = vflow.UITerminal_getNode(to_term)
            to_conn_name = vflow.UITerminal_getName(to_term)
            to_locus = self._children_loci[to_node]
            
            dl_net.add_connection(from_locus, from_conn_name, to_locus, 
              to_conn_name)
            
        
    def _add_of_node(self, name, type):
        # prevent adding duplicates during loading
        if name in self.all_of_nodes.keys():
            return self.all_of_nodes[name]
        else:
            new_node = vflow.new_UINode(self.of_net, name, type, 0.0, 0.0, 1)
            vflow.UINetwork_addNode(self.of_net, new_node)
            
            return new_node
        
                
class TempCompositeLocus(CompositeLocus):
    """Represent the top level of a set of networks.

    The TempCompsiteLocus is a base holder for all other networks inside a
    network diagram. It is analagous to a 'Document' in Overflow.
    """
    def __init__(self, commit_queue):
        """Initialize a new TempCompositeLocus.

        Arguments:
        * commit_queue - A queue item to append automatic saves onto.
        """
        CompositeLocus.__init__(self, None, self)
        
        # set up our environmental variables for overflow
        vflow_envvar = str(OF_LIB)
        
        # load all of the *.def files from Overflow
        toolboxes = os.listdir(BASE_TB_DIR)
        
        for toolbox in toolboxes:
            if (os.path.isdir(os.path.join(BASE_TB_DIR, toolbox)) 
              and toolbox[0] != '.'):
                vflow_envvar = vflow_envvar + ':' + \
                  str(os.path.join(BASE_TB_DIR, toolbox))
                  
        # load all of the Piper *.def files
        piper_defs = os.listdir(BASE_XML_DIR)
        
        for piper_def in piper_defs:
            if (os.path.isdir(os.path.join(BASE_XML_DIR, piper_def)) and
              piper_def[0] != '.' and piper_def != 'PiperUse'):
                vflow_envvar = vflow_envvar + ':' + \
                  str(os.path.join(BASE_XML_DIR, piper_def))
        
        os.environ['VFLOW_PATH'] = vflow_envvar
        
        # load Overflow libs and toolboxes
        vflow.UIDocument_scanDL()
        vflow.UIDocument_loadAllInfo()
        
        # get an Overflow document
        self.of_doc = vflow.new_UIDocument('temp')
        
        self.all_of_networks = {}
        self.commit_queue = commit_queue
        self.id_db = LocusDb()
        
        # create a tempfile for this document for crash recovery
        tempfile.tempdir = TEMP_DIR
        tempfile.template = 'pipertmp'
        
        self.recovery_file = tempfile.mktemp()
                
    def _add_of_network(self, name, type):
        # prevent duplicate adding during loading
        if name in self.all_of_networks.keys():
            return self.all_of_networks[name]
        else:
            new_net = vflow.UIDocument_addNetwork(self.of_doc, name, type)
            self.all_of_networks[name] = new_net
            return new_net
        
    def load_from_file(self, dot_n_file, dl_document):
        """Load a new workspace from a *.n file.
        """
        # set the document to point to the file to load
        vflow.UIDocument_setFullPath(self.of_doc, dot_n_file)
        
        # load the information
        vflow.UIDocument_load(self.of_doc)
        
        # get all of the networks held
        all_networks = vflow.UIDocument_get_networks(self.of_doc)
        
        # create a dictionary of network names and pointers to them
        self.all_of_networks = {}
        for net_pointer in all_networks:
            net_name = vflow.UINetwork_getName(net_pointer)
            self.all_of_networks[net_name] = net_pointer
            
        main_network = None
        # find the MAIN network
        for net_name in self.all_of_networks.keys():
            if net_name == "MAIN":
                main_network = self.all_of_networks[net_name]
                
        if main_network is None:
            raise IOError("File to load did not have a MAIN network defined.")
                
        # add the MAIN network to this one
        main_composite = self.create_locus('composite', "MAIN")
        
        # add a dl network for MAIN to the dl_document
        # and show it in the user interface
        dl_net = dl_document.add_initialized_network(main_composite)
        
        # load the info in the MAIN network
        main_composite._load(main_network, dl_net)
    

        
        
        
        
