"""dl2uiserver.py

This is the middleware server which deals with connections to the user 
interface.

Note: Lots of ideas in this server (and especially the previous parent of 
this) were highly inspired by the AsynchronousServer class 
written by A. Kuchling (akuchlin@mems-exchange.org) and distributed in 
the Remote Microscope package 
(http://www.mems-exchange.org/exchange/software/microscope/). 
Thanks much for the excellent code!
"""
# standard modules
import os

# omniORB
from omniORB import CORBA

# import the corba stubs and skeletons
import modules.piper.UIL2DL
import modules.piper__POA.UIL2DL

# local modules
from compositelocus import TempCompositeLocus
from uicommunication import DlDocument
from blcommunication import BlConnection

# Piper configuration
from piperconfig import *

# --- constants
# BASE_DIR = os.path.join(os.environ['HOME'], 'piper_info', 'temp')

class BaseServer(modules.piper__POA.UIL2DL.DlServer):
    
    def __init__(self, callback_queue, commit_queue, crash_file = None, 
      reload_plugins = 0):
        # a dictionary with all of the connected clients.
        # the keys of the dictionary are connected client objects, and the
        # elments are all of the documents associated with that client.
        self.connected_clients = {}
        
        self.crash_file = crash_file
        
        self.callback_queue = callback_queue
        self.commit_queue = commit_queue
        
        self.clients_to_delete = []
        
        self.reload_plugins = reload_plugins
    
    def set_crash_info(self, info_location):
        """Set crash info that could be loaded by a client."""
        if info_location == self.crash_file:
            return 
        # if we already have crash_info, get rid of it
        if self.crash_file:
            os.remove(self.crash_file)
        
        self.crash_file = info_location
        
    def _remove_crash_info(self):
        """Remove crash information, if we have any."""
        if self.crash_file:
            print "Removing crash info..."
            os.remove(self.crash_file)
            self.crash_file = None  
        
    def addDocument(self, new_client, document_view):
        """Deal with a client request for a new document.
        
        Arguments:
            new_client - A UilClient object defined in the UIL2DL.idl
            file, and contains info about the client, as well as a ping()
            function to call to ensure clients are still alive.
        """
        if new_client is None:
            return      
            
        # create a temporary directory to hold this clients info
        temp_locus = TempCompositeLocus(self.commit_queue)
        
        # create a UiUser object for the user
        new_user = UiUser()

        # connect with the dl
        dl_id, bl_connection = self._connect_with_bl(new_client._get_password())
        new_user.dl_id = dl_id
        new_user.bl_connection = bl_connection
        new_user.callback_queue = self.callback_queue
        new_user.commit_queue = self.commit_queue

        # set the information for the user
        new_user.temp_locus = temp_locus
        new_user.login_info = [new_client._get_username(), 
          new_client._get_password()]

        # get the new document object for the user
        new_doc = DlDocument(new_user, document_view, self, 
          self.reload_plugins)
        
        # if there is crash information, give it to the document
        if self.crash_file:
            new_doc.set_crash_info(self.crash_file)
            self.crash_file = None

        # if the client isn't already in the list of clients, add it
        if new_client not in self.connected_clients.keys():
            self.connected_clients[new_client] = []

        # add the newly created document to be associated with the client
        self.connected_clients[new_client].append(new_doc)

        # get a corba object ready from the document to pass back
        doc_obj = new_doc._this()

        return doc_obj
        
    def _connect_with_bl(self, password):
        """Try to initiate a connection with the bl."""
        dl_id = None
        bl_connection = None
        try:
            # connect with the brokering layer and set the dl_id for the 
            # connection
            print "Connecting with the bl..."
            bl_connection = BlConnection()
            dl_id = bl_connection.setup_connection(password, 'plain')
        # if the server isn't running but the ior file is there
        except CORBA.COMM_FAILURE:
            print 'Brokering layer server not running.'
            print 'Connection with BL not established.'
            bl_connection = None
        # if the ior file is not there.
        except IOError:
            print 'IOR file for Brokering Layer not found.'
            print 'Connection with BL not established.'
            bl_connection = None
            
        return dl_id, bl_connection
        
    def remove_client(self, client):
        """Remove a client from the list of clients being served."""
        self.clients_to_delete.append(client)

    def check_clients(self):
        """Check all of the connected cients to be sure they are still alive.
        
        Calls ping() on each of the clients. If we get a CORBA.COMM_ERROR, we
        assume that the client is dead and unreachable, and clean up everything
        dealing with them.
        
        Returns:
            1 - If there are still clients connected.
            0 - If there are no clients connected.
        """
        # first remove any clients that are scheduled for removal
        for to_delete in self.clients_to_delete:
            for client in self.connected_clients.keys():
                for doc in self.connected_clients[client]:
                    if doc == to_delete:
                        del self.connected_clients[client]
        
        # reset the clients to delete
        self.clients_to_delete = []
        
        for n in self.connected_clients.keys():
            try:
                n.ping()
            except CORBA.COMM_FAILURE:
                for doc in self.connected_clients[n]:
                    temp_file = doc.clean_up(0)
                    if temp_file:
                        self.set_crash_info(temp_file)
                    
                del self.connected_clients[n]
                
        if len(self.connected_clients.keys()) == 0:
            return 0
        else:
            return 1
            
    def do_backups(self):
        """Make temp backup files for all of the clients."""
        for doc_list in self.connected_clients.values():
            for doc in doc_list:
                self.commit_queue.put((doc.save, doc.user.temp_locus.recovery_file))
            
                        
    def clean_up(self, normal = 1):
        for n in self.connected_clients.keys():
            for d in self.connected_clients[n]:
                d.clean_up(normal)
                
            del self.connected_clients[n]
        # delete pid file : 
        pidfilename = os.environ['HOME'] + "/piper_info/pids/DL.pid"
        os.unlink(pidfilename)
        # Now kill the BL, which takes care of killing the PL :
        killBL = "kill `cat " + os.environ['HOME'] + "/piper_info/pids/BL.pid`"
        os.system(killBL);

class UiUser:
    """Represent a user interface user connected to the server.
    
    This provides a way to store all kinds of information associated 
    with a logged in user inteface.
    
    Attributes:
        o temp_locus - The temporary locus holding all of the information
        for the user
        o login_info - A tuple with the users login and password
        o dl_id - The id associated with the connection with the bl.
        o bl_connection - The connection established with the brokering
        layer under the given dl_id.
        o available_programs - A list of all programs available from the
        bl for this user.
        o id_dictionary - The index with all of the locus_id to location
        mappings.
    """
    
    def __init__(self):
        """Initialize a UiUser."""
        self.temp_locus = None
        self.login_info = ['','']
        self.dl_id = None
        self.bl_connection = None
        self.callback_queue = None
        self.commit_queue = None
        
        
    
    
    
