"""dlmain.py

Classes representing the definition layer.

Classes:
    DlMain
"""
# standard modules
import os
import time
import shutil
import sys
import threading
import Queue

# omniORB
from omniORB import CORBA

# local modules
from dl2uilserver import BaseServer

# --- constants
UIL2DL_IOR = os.path.join(os.environ['HOME'], 'piper_info', 'ior', 
 'uil2dl.ior')
BASE_DIR = os.path.join(os.environ['HOME'], 'piper_info', 'temp')
PID_DIR = os.path.join(os.environ['HOME'], 'piper_info', 'pids')

class DlMain:
    """Initiate and manage all of the processes of the definition layer.
    
    This is the main class which should be started for getting the definition 
    layer running. The definition layer is responsible for the following 
    tasks:
        
        * Communicating with (possibly) multiple user interfaces and maintaining 
        XML descriptions of the work flow diagrams in each user interface.
        
        * Permanently storing XML information which can be used in user
        interfaces (such as XML for programs, etc.)
        
        * Communication with the brokering layer to negotiate processing
        of submitted work flow diagrams from the front.
        
        * Communication between definition layers at remote locations to
        allow sharing of nodes.
    """
    
    def __init__(self, reload_plugins):
        """Initialize an instance of the definition layer.
        """
        self.dl_server = None   
        
        # write pid file :
        pid = str(os.getpid())
        pidfilename = os.path.join(PID_DIR, "DL.pid")
        # first make sure the pid directory exists, create it if not
        if not(os.path.exists(PID_DIR)):
            os.makedirs(PID_DIR)
        
        pidfile = open(pidfilename,"w")
        pidfile.write(pid)
        pidfile.close()

        self.callback_queue = Queue.Queue(0)
        self.commit_queue = Queue.Queue(0)
        
        self.shutdown_dl = 0
        
        self.reload_plugins = reload_plugins
        
    def start_ui_communication(self):
        """Initiate a server to communication with user interfaces.
        """
        # see if the dl is already going
        if self._dl_is_running():
            print "The UIL communication layer of the DL is already running."
            print "New DL not started."
            sys.exit(-1)
        
        # check for left over ui info
        crash_file = self._check_for_crash_info()
        
        # initialize the ORB
        self.orb = CORBA.ORB_init([], CORBA.ORB_ID)
        # get the POA and activate the POA Manager
        poa = self.orb.resolve_initial_references("RootPOA")
        poa_manager = poa._get_the_POAManager()
        poa_manager.activate()
        
        self._publish_ior(crash_file)
        
    def start_dl(self):
        """Start up all of the threads which need to run in the dl."""
        t1 = threading.Thread(target = self.run, name = "uil_orb")
        t2 = threading.Thread(target = self.uil_callback, name = "uil_callback")
        t3 = threading.Thread(target = self.do_commits, name = "commit")
        t3.setDaemon(1)
        
        t1.start()
        t2.start()
        t3.start()
        
    def _dl_is_running(self):
        """Check if the uil communication layer of the dl is running.
        
        Returns 1 if it is running, 0 if it is not.
        """
        # if there is no IOR file, it can't be running
        if os.path.exists(UIL2DL_IOR) == 0:
            return 0
            
        # grab our string IOR from the file
        ior_file = open(UIL2DL_IOR, 'r')
        string_ior = ior_file.read()
        ior_file.close()
        
        # now convert the stringified ior into an object reference
        orb = CORBA.ORB_init([], CORBA.ORB_ID)
        dl_connection = orb.string_to_object(string_ior)
        
        try:
            new_doc = dl_connection.addDocument(None, None)
            # if we get here it is running
            return 1
        except CORBA.COMM_FAILURE:
            return 0
        
        # if we got here then my script sucks
        print "Can't detect if the DL is running. That's a bug."
        sys.exit(-1)
                
    def _check_for_crash_info(self):
        """Looks for information indicating a crash.
        """
        if os.path.exists(BASE_DIR):
            possible_files = os.listdir(BASE_DIR)
        else:
            possible_files = []
        
        crash_files = []
        for file in possible_files:
            if file[0:8] == 'pipertmp':
                crash_files.append(os.path.join(BASE_DIR, file))
                
        if len(crash_files) > 1:
            print "Whoops we've got multiple crash files!"
            return crash_files[0]
        elif len(crash_files) == 1: 
            return crash_files[0]
        else:
            return None
        
    def _publish_ior(self, crash_dir):
        """Make the IOR for the server available for clients.
        
        The ior is published to the file:
            $HOME/piper_info/ior/uil2dl.ior
        where $HOME is the home directory of the user running piper.
        """
        # if the directory to hold the ior file doesn't exist, create it
        if os.path.exists(os.path.dirname(UIL2DL_IOR)) == 0:
            os.makedirs(os.path.dirname(UIL2DL_IOR))
            
        # get a server object
        self.dl_server = BaseServer(self.callback_queue, self.commit_queue, 
          crash_dir, self.reload_plugins)
        
        # create an object reference and activate the server
        obj_ref = self.dl_server._this()
        
        # convert the object ref into a string so we can save it to a file
        obj_string = self.orb.object_to_string(obj_ref)
        
        # write the IOR to the file
        ior_file = open(UIL2DL_IOR, "w")
        ior_file.write(obj_string)
        ior_file.close()

        print 'Wrote IOR file ' + UIL2DL_IOR

    def uil_callback(self):
        """Checks for uil_callback items to be sent out and sends them.
        
        The callback items are maintained in a queue to be sent out to
        the user interface. This occurs in a separate thread from the
        recieving to speed things up.
        """
        while not self.shutdown_dl:
            to_send = self.callback_queue.get()
            if to_send:
                if len(to_send) > 1:
                    apply(to_send[0], to_send[1:])
                else:
                    apply(to_send[0])
                    
    def do_commits(self):
        """Perform commits to the XML files when they come in on a queue"""
        while not self.shutdown_dl:
            to_commit = self.commit_queue.get()
            # each commit item has the following format
            # to_commit[0] - The function
            # to_commit[1:] - the arguments to apply
            if to_commit:
                apply(to_commit[0], to_commit[1:])
                
        # when we get here we are getting ready to shut down, but first
        # try and empty the commit_queue
        
        has_items = 1
        while has_items:
            # keep pulling stuff off until we get an empty exception
            try:
                to_commit = self.commit_queue.get(0)
                if to_commit:
                    to_commit[0].acquire()
                    apply(to_commit[1][0], to_commit[1][1:])
                    to_commit[0].release()
            except Queue.Empty:
                has_items = 0
        
    def run(self):
        """Runs the dl in an infinate loop waiting for requests.
        """
        # have an initial sleep to give the clients a chance to 
        # start up
        time.sleep(5)
        continue_loop = 1
        while continue_loop:
            # wait for 10 seconds
            time.sleep(10)
            
            if self.dl_server is None:
                break

            # Call the baseserver to check its clients.
            # If no clients are left, we'll break out of the
            # loop and end the dl
            continue_loop = self.dl_server.check_clients()
            
            # save temporary files for all of the clients
            self.dl_server.do_backups()
            
        # if we got here then we are shutting down the dl and need to 
        # clean everything up
        self.shutdown_dl = 1
        self.callback_queue.put(None)
        print 'shutting down DL...'
        self.dl_server.clean_up(0)
        print 'terminating DL...'
