"""linklocus.py

Provides classes for dealing with a locus that can have links to other loci.

Classes:
    SimpleLinkLocus
"""
# standard modules
import os
import string

# local modules
from locus import Locus

# piper configuration variables
from piperconfig import *

# python wrapping of Overflow UI libs
import vflow

class SimpleLinkLocus(Locus):
    """Represent a locus with connectors.
    """
    def __init__(self, parent_locus, base_locus):
        Locus.__init__(self, parent_locus, base_locus)
        
        self._input_connectors = None
        self._output_connectors = None
        self._connected_inputs = None
        self._connected_outputs = None
            
    def initialize(self, of_node, name):
        """Initialize the information for a locus.

        This function should be called if a locus has already been created,
        but we just need to load the information from it. So this would be
        called when loading a saved file.
        """
        Locus.initialize(self, of_node, name)
            
        # if this is a non-composite type, need to initialize the inputs
        if name not in self.base_locus.all_of_networks.keys() and self.of_node:
            self.of_inputs = vflow.UINode_getInputs(self.of_node)
            self.of_outputs = vflow.UINode_getOutputs(self.of_node)
        
    def create(self, type, name, do_init = 1):
        """Create a new locus of the specified type.

        This function creates a new locus and then calls 'initialize()' to
        initialize the information, if the parameter do_init is set at 1.
        """
        new_node = Locus.create(self, type, name)
        
        if do_init:
            self.initialize(new_node, name)
            
        return new_node
        
    # --- retrieval functions
        
    def get_connector_info(self, type):
        """Retrieve information on all of the locus connectors.

        Returns a 2D list, where each item of the list has the following info:
        list_item[0] - The name of the connector.
        list_item[1] - The type of the connector.
        list_item[2] - The description of the connector
        """
        if type == 'Input':
            if self._input_connectors is None:
                self._input_connectors = []
                
                for input in self.of_inputs:
                    name = string.upper(vflow.UITerminal_getName(input))
                    type = string.lower(vflow.UITerminal_getType(input))
                    description = vflow.UITerminal_getDescription(input)
                    self._input_connectors.append([name, type, description])
                    
            return self._input_connectors
            
        elif type == 'Output':
            if self._output_connectors is None:
                self._output_connectors = []
                
                for output in self.of_outputs:
                    name = string.upper(vflow.UITerminal_getName(output))
                    type = string.lower(vflow.UITerminal_getType(output))
                    description = vflow.UITerminal_getDescription(output)
                    self._output_connectors.append([name, type, description])
                
                #self.file_lock.acquire()
                #self._output_connectors = \
                #  self.storage_locus.get_connector_info(type)
                #self.file_lock.release()
            
            return self._output_connectors
            
        else:
            raise TypeError("Invalid connector type %s passed." % type)
        
    def get_connected_links(self, type):
        """Return information on all of the connected links in the locus.

        Arguments:
        * type - The connector type to retrieve information on. Valid choices
        are 'Input' and 'Output'.

        Returns:
        A 2D list with the following in each item of the list:
        return[0] => name of connector on this locus
        return[1] => locus linked to
        return[2] => name of connector on other locus
        """
        if type == 'Input':
            if self._connected_inputs is None:
                self._connected_inputs = []
                for input in self.of_inputs:
                    if vflow.UITerminal_isConnected(input):
                        this_name = string.upper(
                            vflow.UITerminal_getName(input))
                        connections = vflow.UITerminal_getConnections(input)
                        if len(connections) != 1:
                            raise NotImplementedError \
                              ("Single connector has multiple connections.")
                             
                        # get the output terminal and node
                        output_term = vflow.UILink_getFromTerminal\
                                      (connections[0])
                        output_node = vflow.UITerminal_getNode(output_term)
                        
                        # get the information to append
                        other_name = string.upper(
                            vflow.UITerminal_getName(output_term))
                        other_locus = self._parent._children_loci[output_node]
                        
                        self._connected_inputs.append([this_name, other_locus,
                          other_name])
                        
            return self._connected_inputs   
                               
        elif type == 'Output':
            if self._connected_outputs is None:
                self._connected_outputs = []
                for output in self.of_outputs:
                    if vflow.UITerminal_isConnected(output):
                        this_name = string.upper(
                            vflow.UITerminal_getName(output))
                        connections = vflow.UITerminal_getConnections(output)
                        if len(connections) != 1:
                            raise NotImplementedError \
                              ("Single connector has multiple connections.")
                             
                        # get the input terminal and node
                        input_term = vflow.UILink_getToTerminal(connections[0])
                        input_node = vflow.UITerminal_getNode(input_term)
                        
                        # get the information to append
                        other_name = string.upper(
                            vflow.UITerminal_getName(input_term))
                        other_locus = self._parent._children_loci[input_node]
                        
                        self._connected_outputs.append([this_name, other_locus,
                          other_name])
                
            return self._connected_outputs
                      
        else:
            raise TypeError("Invalid connector type %s passed." % type)
                
    # --- modification functions    
    def add_link(self,  locus_to_link, other_locus_name, this_locus_name, type):
        """Link one of the connectors to another locus connector.

        Arguments:
        * locus_to_link - The locus we are connecting with.
        * other_locus_name - The name of the connector on the other locus with
        which the connection is being made.
        * this_locus_name - The name of the connector on this locus that
        the connection should be made with.
        * type - The type of the connector on this locus ('Input' or 'Output').
        """
        # make sure the names are uppercase
        this_locus_name = string.upper(this_locus_name)
        other_locus_name = string.upper(other_locus_name)
        
        new_link = (this_locus_name, locus_to_link, other_locus_name)
          
        if type == 'Input':
            if self._connected_inputs is None:
                self.get_connected_links(type)
                
            self._connected_inputs.append(new_link)
            
        elif type == 'Output':
            if self._connected_outputs is None:
                self.get_connected_links(type)
                
            self._connected_outputs.append(new_link)
            
        else:
            raise TypeError("Invalid connector type %s passed." % type)
        
    def restore_link(self, link_type, link_id):
        """Reset a linked connector to be unlinked.

        Arguments:
        * link_type - The type of the link to restore ('Input' or 'Output').
        * link_id - The name of the link which should be restored.
        """
        # make sure the link_id is uppercase
        link_id = string.upper(link_id)
        
        if link_type == 'Input':
            if self._connected_inputs is None:
                self.get_connected_links(link_type)
                
            for link in self._connected_inputs:
                if link[0] == link_id:
                    to_remove = link
                    
            self._connected_inputs.remove(to_remove)
            
        elif link_type == 'Output':
            if self._connected_outputs is None:
                self.get_connected_links(link_type)
                
            for link in self._connected_outputs:
                if link[0] == link_id:
                    to_remove = link
                    
            self._connected_outputs.remove(to_remove)
            
        else:
            raise TypeError("Invalid connector type %s passed." % link_type)     
        
    def remove_connector(self, type, name):
        """Remove a connector from the locus.

        Arguments:
        * type - The type of the connector to remove ('Input' or 'Output')
        * name - The unique name of the connector to be removed.
        """
        # be sure the name is uppercase
        name = string.upper(name)
        
        if type == 'Input':
            # first remove the connector from the list of all connectors
            if self._input_connectors is None:
                self.get_connector_info(type)
                
            for conn in self._input_connectors:
                if conn[0] == name:
                    to_remove = conn
                    
            self._input_connectors.remove(conn)
            
            # if the connector is connected, remove it from the connected list
            if self._connected_inputs is None:
                self.get_connected_links(type)
                
            to_remove = None
            for link in self._connected_inputs:
                if link[0] == name:
                    to_remove = link
                    
            if to_remove:
                self._connected_inputs.remove(link)
            
        elif type == 'Output':
            # first remove the connector from the list of all connectors
            if self._output_connectors is None:
                self.get_connector_info(type)
                
            for conn in self._output_connectors:
                if conn[0] == name:
                    to_remove = conn
                    
            self._output_connectors.remove(conn)

            # if the connector is connected, remove it from the connected list
            if self._connected_outputs is None:
                self.get_connected_links(type)
                
            to_remove = None
            for link in self._connected_outputs:
                if link[0] == name:
                    to_remove = link
                    
            if to_remove:
                self._connected_outputs.remove(link)
            
        else:
            raise TypeError("Invalid connector type %s passed." % type) 
              
    def add_connector(self, conn_type, name, link_type, description):
        """Add a new connector to the locus.

        Arguments:
        * conn_type - The type of the new connector ('Input' or 'Output')
        * name - The name of the new connector, which must be unique.
        * link_type - A string specifying the type of information the connector
        will support.
        * description - A string providing information about the function of
        the connector.
        """
        # make sure the link_type is lowercase
        link_type = string.lower(link_type)

        # make sure the name is uppercase
        name = string.upper(name)

        # first update our local information on available connectors
        
        new_conn = (name, link_type, description)
        
        if conn_type == 'Input':
            if self._input_connectors is None:
                self.get_connector_info(type)
                
            self._input_connectors.append(new_conn)
            
        elif conn_type == 'Output':
            if self._output_connectors is None:
                self.get_connector_info(type)
                
            self._output_connectors.append(new_conn)
        
        else:
            raise TypeError("Invalid connector type %s passed." % conn_type)

        # TODO: add the new connector information to the overflow info
        if self.of_node:
            pass



            

                
        
        
        
        
                
        
        
