"""locus.py

This is the main overall class defining a locus and providing the ability 
to manipulate it's xml files and get information from it.

Classes:
    Locus
"""
# standard modules
import os
import string

# Overflow wrapped for python
import vflow

class Locus:
    """Base class for representing a in-memory Locus.
    
    This class stores the information for a locus in temporary memory and
    accesses the PersistentLocus class to save and retrieve information
    from disk storage.
    """
    def __init__(self, parent_locus, base_locus):
        self._net_name = None
        self._type = None
        self._parent = parent_locus
        self._position = None
        self._description = None
        self._parameters = None # name type value description
        
        self.base_locus = base_locus
        self.of_net = None
        
        self.of_node = None
        self.of_params = None
        
    def initialize(self, of_node, name = None):
        """Initialize a new locus with the given overflow node object.

        After initializing the new locus, it is added to the list of loci
        contained within its parent locus.
        """
        self.id_name = name
        self.of_node = of_node
        
        if self.of_node:
            # get the parameters for the new node
            self.node_param = vflow.UINode_getParameters(self.of_node)
            self.of_params = vflow.UINodeParameters_get_textParams\
                             (self.node_param)
        
            # set the position information
            x_pos, y_pos = vflow.UINode_getPos(self.of_node)
            self.set_position(x_pos, y_pos)
            
            
            # add the node to the list of nodes in the parent composite
            if self._parent._children_loci is None:
                self._parent.get_children_loci()
                
            self._parent._children_loci[self.of_node] = self
                
    def create(self, type, name):
        """Create and add a new overflow node object of the given type."""
        new_name = name
        if self._parent and self._parent.of_net:
    
            if (type == 'composite' or type == 'container'
                or type == 'iterator'):
                name = 'node_' + new_name
                type = new_name
            else:
                name = new_name
                
            new_node = self._parent._add_of_node(name, type)
              
            return new_node
            
        # if we didn't create anything  
        return None
            
    def __cmp__(self, to_compare):
        """A cheesy compare for locus objects."""
        if self.id_name == to_compare.id_name:
            return 1
        else:
            return 0    
        
    # --- retrieval functions
            
    def get_net_name(self):
        """Return a name for the locus which is unique within a network."""
        return self.id_name
        
    def get_description(self):
        """Return the description of the locus."""
        if self._description is None:
            if self.of_node:
                self._description = vflow.UINode_getDescription(self.of_node)
            else:
                self._description = "No description available."
            
        return self._description
        
    def get_parameters(self):
        """Return all of the parameters of the locus.

        Returns:
        A 2D list where each list item has the following info:
        * list_item[0] - the name of the parameter.
        * list_item[1] - the type of the parameter
        * list_item[2] - the value of the parameter
        * list_item[3] - a description of the parameter
        """ 
        if self._parameters is None:
            self._parameters = {}
            
            if self.of_params:
            
                for of_param in self.of_params:
                    name = string.upper(vflow.ParameterText_get_name(of_param))
                    value = vflow.ParameterText_get_value(of_param)
                    type = string.lower(vflow.ParameterText_get_type(of_param))
                    description = vflow.ParameterText_get_description(of_param)
                
                    self._parameters[of_param] = [name, type, value,
                                                  description]
            
        return self._parameters.values()
        
    def get_type(self):
        """Return the type of the locus."""
        if self._type is None:
            
            self._type = vflow.UINode_getType(self.of_node)
            
            if self.base_locus:
                # if it is a network we need to find the type of network
                if self._type in self.base_locus.all_of_networks.keys():
                    the_network = self.base_locus.all_of_networks[self._type]
                    num_type = vflow.UINetwork_getType(the_network)
                    if num_type == 0:
                        self._type = 'composite'
                    elif num_type == 1:
                        self._type = 'iterator'
                    else:
                        raise NotImplementedError\
                              ("Cannot get network type %s." % num_type)
                
            
        return self._type
            
    def get_parent(self):
        """Retrieve the parent locus of this locus.
        """
        if self._parent is None:
            raise NotImplementedError("Whoops. Can't get a parent for this!")
            
        return self._parent
        
    def get_position(self):
        """Return the position of the locus (x, y).
        """
        if self._position is None:
            self._position = vflow.UINode_getPos(self.of_node)
        
        return self._position
            
    # --- modification functions
        
    def set_parameter_type(self, param_name, param_type):
        """Set a new type for a particular parameter.

        Arguments:
        * param_name - The unique name of the parameter.
        * param_type - The new type to set the parameter as.
        """
        # make sure the parameter type is lowercase
        param_type = string.lower(param_type)

        # make sure parameter name is uppercase
        param_name = string.upper(param_name)

        if self._parameters is None:
            self.get_parameters()
            
        param_changed = 0
        changed_key = None
        for param_key in self._parameters.keys():
            if self._parameters[param_key][0] == param_name:
                self._parameters[param_key][1] = param_type
                changed_key = param_key
                
        if changed_key is None:
            raise NameError("Parameter with name %s was not found."
                            % param_name)
        
        # change the info in the Overflow stuff
        vflow.ParameterText_set_type(changed_key, param_type)
        
    def set_parameter_info(self, param_name, param_value):
        """Set the value for a parameter.

        Arguments:
        * param_name - The unique name of the parameter to modify.
        * param_value - The new value for the parameter.
        """
        # make sure the parameter name is uppercase
        param_name = string.upper(param_name)
        
        if self._parameters is None:
            self.get_parameters()
            
        param_changed = 0
        changed_key = None
        for param_key in self._parameters.keys():
            if self._parameters[param_key][0] == param_name:
                self._parameters[param_key][2] = param_value
                changed_key = param_key
            
        if changed_key is None:
            raise NameError("Parameter with name %s was not found."
                            % param_name)
            
        # change the info in the Overflow stuff
        vflow.ParameterText_set_value(changed_key, param_value)
          
    def add_parameter(self, name, type, description):
        """Add a new parameter to the locus.

        Arguments:
        * name - A unique name for the new parameter.
        * type - The type of the new parameter.
        * description - A description of the use of the parameter.
        """
        # make sure paramter name is uppercase
        name = string.upper(name)
        
        if self._parameters is None:
            self.get_parameters()   
            
        for param in self._parameters.values():
            if param[0] == name:
                raise KeyError("Passed name %s is not a unique key." % name)

        # make sure the type is lowercase
        type = string.lower(type)
                
        # add a new parameter to the Overflow info
        new_param = vlow.UINodeParameters_addParameterText(self.node_param,
          name, type, "", description)
            
        self._parameters[new_param] = [name, type, "", description]
        self.of_params.append(new_param)
            
    def remove_parameter(self, name):
        """Remove a parameter from the locus parameter.

        Arguments:
        * name - The unique name of the parameter to remove.
        """
        # ensure the parameter name is uppercase
        name = string.upper(name)
        
        if self._parameters is None:
            self.get_parameters()
            
        to_remove_key = None
        for param_key in self._parameters.keys():
            if self._parameters[param_key][0] == name:
                to_remove_key = param_key
                
        if to_remove_key is None:
            raise KeyError("A parameter with name %s was not found." % name)
            
        # remove the parameter from the overflow info
        vflow.UINodeParameters_removeParameterText(self.node_param, name)
                
        del self._parameters[to_remove_key]
        
    def set_position(self, x, y):
        """Set the position of the locus."""
        self._position = (x, y)
        
        # set the position in of
        vflow.UINode_setPos(self.of_node, x, y)
        
    
                


                
        
        
        
