"""locusid.py

Classes for dealing with LocusIDs and converting between ids and physical 
locations of xml files (in the directory structure).

classes:
    LocusId
"""
# standard modules
import os
import os.path
import random
import string
import shutil

# configuration variables
from piperconfig import *

class LocusDb:
    """Deal with ids and names for all of the loci in a document."""
    
    def __init__(self):
        """Create a new repository of locus id names."""
        # readible locus names that are unique within the document
        self.id_names = []
        
    def is_registered(self, id_name):
        """Verify if the passed id name is registered in the repository."""
        if id_name in self.id_names:
            return 1
        else:
            return 0
        
    def get_new_id(self, type):
        """Add a new locus and name to the repository.
        
        The id name will be assigned to a default of typeNUM, where NUM is 
        a unique number.
        
        Returns:
            o the unique key name assigned to the locus.
        """
        cur_num_list = []
        for id in self.id_names:
            # if the base of the id name is equal to type
            if id[0:len(type)] == type:
                # append the number to the list of taken numbers
                cur_num_list.append(string.atoi(id[len(type):]))
                
        if len(cur_num_list) == 0:
            new_num = 1
        else:
            cur_num_list.sort()
            high_num = cur_num_list[len(cur_num_list) - 1]
            new_num = high_num + 1
            
        new_id = type + str(new_num)
        
        # do a sanity check to make sure this is working
        if new_id in self.id_names:
            raise NotImplementedError("Damn, get_new_id does not work.")
            
        self.id_names.append(new_id)
        
        return new_id
        
    def add_id(self, new_name):
        """Register a new id with the database.
        
        Raises:
            KeyError - If the name already exists in the database.
        """
        if new_name in self.id_names:
            raise KeyError("The passed id %s is in use." % new_name)
            
        self.id_names.append(new_name)
        
    def change_item_name(self, old_name, new_name):
        """Change the name of the item from the old to the new.
        
        Raises:
            KeyError -> If the new passed name is not unique.
        """
        if new_name in self.id_names:
            raise KeyError("The passed id %s is not unique." % new_name)
            
        self.id_names.remove(old_name)
        self.id_names.append(new_name)
