# standard modules
import string
import os

# constants
from piperconfig import *

class PiperDbConnect:
    """
    This class deals with database connectivity in Piper. It is an attempt
    to provide a "database" independent way of dealing with queries, etc.
    within Piper.
    For now, all database interfaces that will be supported will need to 
    conform to the python database API spec version 2.0. 
    
    Currently supported databases:
        MySQL
        PostgreSQL
        Gadfly -- Note: this uses the constant gadflyDirectory, defined
        in main.library.modules.const.py, as the directory to store
        the gadfly info in.
    """
    
    def __init__(self, dbtype = '', host_name = '', user_name = '', password = '',
                 port_num = '', socket = ''):
        """
        The class is initialized with the parameters required to connect
        to a database:
            
            - DBtype: The database type. Currently Supported:
                --MySQL
            - host_name: The host to connect to.
            - user_name: The user to connect as.
            - password: The password to use.
            - port_num: TCP/IP port to connect to (if not set, will be the
              default port)
            - socket: location of unix_socket to use (if not set, will
              use TCP)
              
        A database connection will then be established, and will be 
        ready for more queries.
        
        TO DO: Need to figure out what kind of errors this can generate
        and handle them in an appropriate manner.
        """
        # Build a dictionary of keywords/arguments to connect with, 
        # using whatever passed parameters are not ''.
        # the dictionary is used to connect to MySQL.
        # For connecting to PostgreSQL, we need the name directly.
        connect_info = {}
        if (host_name != ''):
            connect_info['host'] = host_name
            self.host_name = host_name
        else:
            self.host_name = None
        if (user_name != ''):
            connect_info['user'] = user_name
            self.user_name = user_name
        else:
            self.user_name = None
        if (password != ''):
            connect_info['passwd'] = password
            self.password = password
        else:
            self.password = None
        if (port_num != ''):
            connect_info['port'] = port_num
            self.port_num = port_num
        else: 
            self.port_num = None
        if (socket != ''):
            connect_info['unix_socket'] = socket
            self.socket = socket
        else:
            self.socket = None
    
        # find the right interface to import and connect to
        if (dbtype != ''):
            if (string.lower(dbtype) == 'mysql'):
                import MySQLdb
                self.connection = apply(MySQLdb.connect, (), connect_info)
                # get a cursor to do everything with
                self.cursor = self.connection.cursor()
            if (string.lower(dbtype) == 'postgresql'):
                import pgdb
                # initially connects to the template1 database, so we can get 
                # the other databases--is there a better way to do this?
                self.connection = pgdb.connect(None, self.user_name, 
                  self.password, self.host_name, 'template1')
                # get a cursor to do everything with
                self.cursor = self.connection.cursor()
            if (string.lower(dbtype) == 'gadfly'):
                import gadfly
                self.connection = gadfly.gadfly()
        
        # keep track of what type of database we are connecting to
        self.type = string.lower(dbtype)
        
    def get_databases(self):
        """
        Returns all of the available databases in the current connection
        as a list.
        """
        if (self.type == 'mysql'):
            getDBsql = "SHOW DATABASES"     
            
        elif (self.type == 'postgresql'):
            getDBsql = "select datname from pg_database"
        
        # for gadfly we don't do an SQL select to get the databases
        # so we'll just look in the directory with the storage info
        # and find the database names and return these
        elif (self.type == 'gadfly'):
            dbs = list(())
            files = os.listdir(gadflyDirectory)
            for n in files:
                # if the file is a gadfly file, save the name
                # as a database name
                name, extension = os.path.splitext(os.path.basename(n))
                if (extension == '.gfd'):
                    dbs.append(name)
                    
            return dbs
            
        print 'type:', self.type
        # if we're not using gadfly, then we need to execute the sql
        # for other databases
        self.cursor.execute(getDBsql)
        returnval = self.cursor.fetchall()
        databases = list(())
        # convert the returned 2D array into a single list
        for n in returnval:
            databases.append(n[0])
            
        # For PostgreSQL, we need to get rid of the template1 default 
        # database and show only user created databases.    
        if (self.type == 'postgresql'):
            self.t1Deleted = 0
            for n in range(len(databases)):
                if (self.t1Deleted == 0):
                    if (databases[n] == 'template1'):
                        del databases[n]
                        self.t1Deleted = 1
            
        return databases
        
    def get_tables(self, database):
        """
        When passed the name of a database, retrieves all of the tables
        from the database and returns them as a list.
        """
        if (self.type == 'mysql'):
            # change the database
            changeDBsql = "USE " + database
            self.cursor.execute(changeDBsql)
            # show the table info
            getTablesql = "SHOW TABLES"
            
        elif (self.type == 'postgresql'):
            # close the current database connection and open a
            # connection to the new database
            self.connection.close()
            import pgdb
            self.connection = pgdb.connect(None, self.user_name, self.password, self.host_name, database) 
            # get the new cursor
            self.cursor = self.connection.cursor()
            
            # the table info sql
            getTablesql = "select tablename from pg_tables where not tablename ~ 'pg_.*'"
            
        elif (self.type == 'gadfly'):
            # connect to the database
            import gadfly
            self.connection = gadfly.gadfly(database, gadflyDirectory)
            self.cursor = self.connection.cursor()
            getTablesql = "select * from __table_names__"
            
        self.cursor.execute(getTablesql)
        
        returnval = self.cursor.fetchall()
        # take care of the special case for cleaning up the returned 
        # info from a gadfly table select
        if (self.type == 'gadfly'):
            tables = list(())
            for n in returnval:
                # just get the unique tables that have been created
                if (n[0] == 0):
                    tables.append(n[1])
        else:
            tables = list(())
            # convert the returned 2D array into a single list
            for n in returnval:
                tables.append(n[0])
        
        return tables
        
    def get_table_columns(self, database, table):
        """
        When passed a database and table, this returns the names of
        each of the columns in the table as a list.
        """
        if (self.type == 'mysql'):
            # first change to the new database
            changeDBsql = "USE " + database
            self.cursor.execute(changeDBsql)
            # the sql for getting the column info
            getColumnsql = "DESCRIBE " + table
        
        elif (self.type == 'postgresql'):
            # close the current database connection and open a
            # connection to the new database
            self.connection.close()
            import pgdb
            self.connection = pgdb.connect(None, self.user_name, self.password, self.host_name, database) 
            # get the new cursor
            self.cursor = self.connection.cursor()
            
            # column select sql
            getColumnsql = "SELECT attname, typname FROM pg_attribute, pg_type WHERE typrelid=attrelid AND typname = '" + table + "'"
        
        elif (self.type == 'gadfly'):
            # connect to the database
            import gadfly
            self.connection = gadfly.gadfly(database, gadflyDirectory)
            self.cursor = self.connection.cursor()
            getColumnsql = "SELECT * FROM __columns__"
            
        self.cursor.execute(getColumnsql)
        
        returnval = self.cursor.fetchall()
        cols = list(())
        
        # if we have a gadfly database the output data is returned in a
        # "non-normal" way, so we need to extract the columns we need 
        # and return these
        if (self.type == 'gadfly'):
            cols = list(())
            for n in returnval:
                if (n[1] == table):
                    cols.append(n[0])
                    
        # if not a gadfly db, then deal with things in the normal way
        else:
            # go through the returned 2D array of information and get
            # only the column name info
            for n in returnval:
                cols.append(n[0])
        
            # the postgresql column info select command returns six extra
            # items before the table names, so we have to chop this off
            if (self.type == 'postgresql'):
                del cols[:6]
                # reorder cols because the column names are
                # backwards for postgresql
                cols.reverse()
                    
        return cols
        
    def get_fields(self, database, table):
        """
        Selects all of the fields from the passed database and table
        (ie. SELECT * FROM table) and returns the information as a tuple, 
        with each element of the tuple being a tuple containing the info
        in a row of the database.
        """
        if (self.type == 'mysql'):
            # change to the database
            changeDBsql = "USE " + database
            self.cursor.execute(changeDBsql)
        
        elif (self.type == 'postgresql'):
            # close the current database connection and open a
            # connection to the new database
            self.connection.close()
            import pgdb
            self.connection = pgdb.connect(None, self.user_name, self.password, self.host_name, database) 
            # get the new cursor
            self.cursor = self.connection.cursor()
                
        elif (self.type == 'gadfly'):
            # connect to the database
            import gadfly
            self.connection = gadfly.gadfly(database, gadflyDirectory)
            self.cursor = self.connection.cursor()
            
        
        getInfosql = "SELECT * FROM " + table + " ORDER BY 1 DESC"
        
        self.cursor.execute(getInfosql)
        
        returnval = self.cursor.fetchall()
        
        print 'return:', returnval
        
        return returnval
        
    def is_database(self, to_check):
        """
        Determines if the passed file is a database in the current 
        database connection. If it is, 1 (TRUE) is returned, otherwise
        0 (FALSE) is returned.
        """
        dbs = self.get_databases()
        isadb = 0
        for n in dbs:
            if (n == string.strip(to_check)):
                isadb = 1
                
        return isadb
        
    def get_table_db(self, to_check):
        """
        Determines if the passed file is a table in the current 
        database connection. If it is, the name of the database which
        contains it is returned, otherwise, None is returned.
        """

        tableDB = None      
        # check each database for the folder
        dbs = self.get_databases()
        for d in dbs:
            tables = self.get_tables(d)
            for t in tables:
                if (t == string.strip(to_check)):
                    tableDB = d
        
        return tableDB
                
                
            
        