"""uicommunication.py

Contains classes for dealing with talking directly with the User Interface.
This module implements the CORBA interface described in UIL2DL.idl. 

classes:
    DlDocument
    DlNetwork
    DlLocus
    DlConnector
    DlParameter
    DlProcess
"""

# standard modules
import string
import os

# omniORB
from omniORB import CORBA

# import the corba stubs and skeletons
import modules.piper.UIL2DL
import modules.piper__POA.UIL2DL

# local modules
from compositelocus import TempCompositeLocus
from compositelocus import CompositeLocus
from linklocus import SimpleLinkLocus

# Overflow 
import vflow

# --- constants
SUBMITTED_XML = os.path.join(os.environ['HOME'], 'piper_info', 'storage',
                             'submitted.xml')


# --- some helpful functions for dealing with enums 

def param_type_string_to_corba(type):
    """Convert the string name for a parameter type into a corba enum.
        
    If the passed type is not a valid type, then it will be set as
    ANY, the catch-all type.
    """
    type = string.lower(type)
    if type == 'value':
        return modules.piper.UIL2DL.Parameter.P_VALUE
    elif type == 'string':
        return modules.piper.UIL2DL.Parameter.P_STRING
    elif type == 'int':
        return modules.piper.UIL2DL.Parameter.P_INT
    elif type == 'float':
        return modules.piper.UIL2DL.Parameter.P_FLOAT
    elif type == 'bool':
        return modules.piper.UIL2DL.Parameter.P_BOOL
    elif type == 'subnet_param':
        return modules.piper.UIL2DL.Parameter.P_SUBNET_PARAM
    elif type == 'any':
        return modules.piper.UIL2DL.Parameter.P_ANY
    else:
        return modules.piper.UIL2DL.Parameter.P_ANY
            
def param_type_corba_to_string(type):
    """Convert the corba enum into a string name for the parameter type."""
    if type == modules.piper.UIL2DL.Parameter.P_VALUE:
        return 'value'
    elif type == modules.piper.UIL2DL.Parameter.P_STRING:
        return 'string'
    elif type == modules.piper.UIL2DL.Parameter.P_INT:
        return 'int'
    elif type == modules.piper.UIL2DL.Parameter.P_FLOAT:
        return 'float'
    elif type == modules.piper.UIL2DL.Parameter.P_BOOL:
        return 'bool'
    elif type == modules.piper.UIL2DL.Parameter.P_SUBNET_PARAM:
        return 'subnet_param'
    elif type == type == modules.piper.UIL2DL.Parameter.P_ANY:
        return 'any'
    else:
        return 'any'
            
def conn_type_string_to_corba(type):
    """Convert the string name for a connector type into a corba enum.
        
    If the passed type is not a valid type, then it will be set as
    ANY, the catch-all type.
    """
    type = string.lower(type)
    if type == 'value':
        return modules.piper.UIL2DL.Connector.P_VALUE
    elif type == 'stream':
        return modules.piper.UIL2DL.Connector.P_STREAM
    elif type == 'object':
        return modules.piper.UIL2DL.Connector.P_OBJECT
    elif type == 'condition':
        return modules.piper.UIL2DL.Connector.P_CONDITION
    elif type == 'string':
        return modules.piper.UIL2DL.Connector.P_STRING
    elif type == 'any':
        return modules.piper.UIL2DL.Connector.P_ANY
    else:
        return modules.piper.UIL2DL.Connector.P_ANY
        
def conn_type_corba_to_string(type):
    """Convert the corba type of a connector into the corresponding string."""
    if type == modules.piper.UIL2DL.Connector.P_VALUE:
        return 'value'
    elif type == modules.piper.UIL2DL.Connector.P_STREAM:
        return 'stream'
    elif type == modules.piper.UIL2DL.Connector.P_OBJECT:
        return 'object'
    elif type == modules.piper.UIL2DL.Connector.P_CONDITION:
        return 'condition'
    elif type == modules.piper.UIL2DL.Connector.P_STRING:
        return 'string'
    elif type == modules.piper.UIL2DL.Connector.P_ANY:
        return 'any'
    else:
        return 'any'
            
# --- classes

class DlDocument(modules.piper__POA.UIL2DL.Document):
    
    def __init__(self, user, uil_document_view, dl_server, 
      reload_plugins = 0):
        self.dl_server = dl_server
        self.user = user
        
        # maintain a list of all Loci selected in a document
        self.selected_loci = []
        
        # a list of all networks within this document
        self.networks = []
        
        # a callback to add info about a document
        self.uil_document_view = uil_document_view
        
        # location of crash information that can be loaded
        self.crash_file = None
        
        # whether or not we need to reload the plugins
        self.reload_plugins = reload_plugins
        
        # create our initial recovery file
        self.save(self.user.temp_locus.recovery_file)
        
    # --- corba functions
    def _get_heldNetwork(self):
        """Return all the networks held by this document."""
        net_ob_list = []
        for net in self.networks:
            net_ob_list.append(net._this())
            
        return net_ob_list[0]
    
    def getPlugins(self):
        """Return the locus plugins which are available to the uil.
        
        For now, I am just going to return a list of all plugins with xml
        files and later we'll need to restrict this so that we only show
        loaded plugins, and maybe we'll even restrict plugin availability
        based on the submitted name and password.
        """
 
        # Commented out, DL no longer communicates to PL directly - Jarl 26-7-2001
        of_plugins = \
          vflow.UIDocument_getAvailableNodes(self.user.temp_locus.of_doc)
        
        locus_list = []
        for of_plugin in of_plugins:
            # print "Parsing Oveflow string %s" % of_plugin
            category, name = string.split(of_plugin, '***')
            locus_list.append((category, name))
        
        # if we are connected with the bl, get the available plugins
        # and only return those to the user
        if self.user.bl_connection:
            # get the available programs by id
            available_ids = self.user.bl_connection.get_program_ids()
            # TODO -> convert the ids to program names
            available_plugins = available_ids
            for element in available_plugins:
                uri, category_name = string.split(element,':::')
                # print "parsing PL string %s" % category_name
                category, name = string.split(category_name, '***')
                locus_list.append((category, name))

            # good_plugins = [] 
            good_plugins = locus_list
            # only return those plugins that are available
            for locus in locus_list:
                if locus[1] in available_plugins:
                    good_plugins.append(locus)
                
        # otherwise, just return everything
        else:
            good_plugins = locus_list
        
        # now convert the list of locus plugins into the right form
        # to pass back.
        struct_locus_list = []
        for locus in good_plugins:
            new_struct_locus = modules.piper.UIL2DL.Plugin(locus[0], locus[1])
            struct_locus_list.append(new_struct_locus)
            
        return struct_locus_list
        
    def addPlugin(self, plugin_to_add):
        pass
            
    def _get_selectedLoci(self):
        return self.selected_loci
        
    def clearSelectedLoci(self):
        self.selected_loci = []
        
    def setSelectedLoci(self, locus):
        self.selected_loci = [locus]
        
    def addSelectedLoci(self, locus):
        self.selected_loci.append(locus)
        
    def setNetwork(self, type):
        """Create a new Network within the document.
        """
        valid_networks = ['composite', 'iterator']
        if type not in valid_networks:
            raise modules.piper.UIL2DL.TypeError \
              ("Invalid type %s passed for creating a Network." % type) 
        
        new_net = DlNetwork(self.user, type, None, None, 1)
        self.networks = []
        self.networks.append(new_net)
        
        # call the user interface to add the network
        new_view = self.uil_document_view.showNetwork(new_net._this())
        new_net.set_view_callback(new_view)
            
    def run(self):
        """Submit a document to the brokering layer to be evaluated.
        
        The document chooses which loci to submit based on the selected loci
        for the document. It is assumed that all of the selected loci are
        the final items to be evaluated in a work flow diagram, and so
        the dl uses a pull model to continually grab parent loci to this
        locus until we have all of the loci necessary to get the final
        results needed.
        
        This document is then built into an XML file and submitted to the bl.
        """
        #if self.user.bl_connection is None:
        #   raise modules.piper.UIL2DL.ProcessingError \
        #     ("Not connected to Brokering Layer. Can't submit document")
        
        if len(self.selected_loci) == 0:
            raise modules.piper.UIL2DL.ProcessingError \
              ("No items selected to be submitted.")

        # the net termini that were added to run
        add_net_terms = []
        
        # All of the unconnected outputs in selected loci will be mapped
        # to the top level network so that we can "pull" from these.
        # TODO: Should eventually just pull from selected connectors.
        try:
            selected_loci_obj = []
            # get the locus objects for the selected loci
            for remote_locus in self.selected_loci:
                print 'remote_locus:', remote_locus
                print 'self.networks', self.networks
                print 'self.networks[0],', self.networks[0]
                print 'self.networks[0].held_loci', self.networks[0].held_loci
                for dl_locus in self.networks[0].held_loci:
                    print "dl_locus:", dl_locus
                    if dl_locus._this()._is_equivalent(remote_locus):
                        local_locus = dl_locus.locus

                selected_loci_obj.append(local_locus)
                
            added_net_terms = self.networks[0].locus.map_locus_outputs \
              (selected_loci_obj)
        except LookupError:
            raise modules.piper.UIL2DL.ProcessingError \
              ("No free outputs were found in the selected items.")

        # create the directory to save in if it doesn't already exist
        if not(os.path.exists(os.path.dirname(SUBMITTED_XML))):
            os.makedirs(os.path.dirname(SUBMITTED_XML))
              
        # create a document for submission to the bl
        
        # TODO: Right now we just send everything in the network description,
        # while we should be just sending everything connected the selected
        # nodes. Although this might be okay since we just pull from the selected
        # loci. Hmmm...
        vflow.UIDocument_setFullPath(self.user.temp_locus.of_doc, SUBMITTED_XML)
        vflow.UIDocument_save(self.user.temp_locus.of_doc)

        # now submit the document to the brokering layer
        if self.user.bl_connection:
            submit_file = open(SUBMITTED_XML, 'r')
            xml_to_submit = submit_file.read()
            submit_file.close()
            # self.user.bl_connection.rep_ref.upload(xml_to_submit,1,"password")
            self.user.bl_connection.submit_representation(xml_to_submit)

        # otherwise we can run it directly through overflow wrappers
        else:
            vflow.UIDocument_run(self.user.temp_locus.of_doc)

        # remove the net_termini we mapped
        self.networks[0].locus.unmap_net_terms(added_net_terms)
           
    def save(self, filename):
        """Save the current document into the passed filename."""
        # create the directory to save in if it doesn't already exist
        if not(os.path.exists(os.path.dirname(filename))):
            os.makedirs(os.path.dirname(filename))
        
        vflow.UIDocument_setFullPath(self.user.temp_locus.of_doc, filename)
        vflow.UIDocument_save(self.user.temp_locus.of_doc)
         
    def close(self):
        # first close the connection with the dl if it exists
        if self.user.bl_connection:
            self.user.bl_connection.release_connection()
        
        # now clean up local junk
        self.clean_up(1)        
        
        # # call the server to remove this client
        self.dl_server.remove_client(self)
        
    def loadCrashInfo(self):
        """Load a directory of crash information into the current document."""
        if self.crash_file is None:
            raise modules.piper.UIL2DL.LookupError \
              ("No crash information available.")
            
        # disable crash recovery for now
        else:
            try:
                self.loadDocument(self.crash_file)
            except modules.piper.UIL2DL.IOError:
                self.crash_file = None
                raise modules.piper.UIL2DL.LookupError \
                  ("No crash information available.")
                  
                  
            # remove the crash file now that we've loaded from it
            os.remove(self.crash_file)
            self.crash_file = None
          
    def loadDocument(self, file):
        if not(os.path.exists(file)):
            raise modules.piper.UIL2DL.IOError \
              ("File to load could not be found in the filesystem.")
              
        # get rid of the current directory
        self.clean_up(1)
            
        # get a new directory
        del(self.user.temp_locus)
        self.user.temp_locus = TempCompositeLocus(self.user.commit_queue)
        
        self.user.temp_locus.load_from_file(file, self)         
        
        # write the crash info for this new file
        self.save(self.user.temp_locus.recovery_file)
        
    def loadDocumentUrl(self, url):
        pass
        
        
    # --- local functions
    def set_crash_info(self, info_location):
        """Set a directory of crash information that can be loaded."""
        if info_location == self.crash_file:
            return 
        # if we already have crash_info, get rid of it
        if self.crash_file:
            os.remove(self.crash_file)
            
        self.crash_file = info_location
        
    def clean_up(self, normal = 1):
        """Clean up the files associated with the document.
        
        Arguments:
            normal - When set at 1, we do a normal clean-up and delete all 
            all of the files. A normal = 0 argument should be called in 
            the case of a crash (which the middle assumes if Document.close()
            is not called).
        """
        if normal == 1:
            # remove the temporary file for the document
            os.remove(self.user.temp_locus.recovery_file)
            
        else:
            if os.path.exists(self.user.temp_locus.recovery_file):
                print "Crash detected - Not deleting document info."
                # save the most recent copy of this file
                self.save(self.user.temp_locus.recovery_file)
                
                return self.user.temp_locus.recovery_file
            
            return None
            
    def add_initialized_network(self, init_composite):
        """Add a new network with an initialized composite locus.
        
        Returns:
            The created DlNetwork
        """
        dl_net = DlNetwork(self.user, 'composite', None, init_composite, 1)
        new_net_view = self.uil_document_view.showNetwork(dl_net._this())
        dl_net.set_view_callback(new_net_view)
        self.networks = []
        self.networks.append(dl_net)
        
        return dl_net

    # --- callback functions
    def send_output(self, node_name, string_info):
        print 'Python callback recieved:'
        print '  Node Name', node_name
        print '  Node Info', string_info
        
    
    """
    def load_info(self, is_document = 0):
        Load the information to populate the DlDocument from xml files.
        
        This function allows you to create info available for a UI from
        stored files or crash files (as opposed to being created directly
        by the user interface).
        This will call load_info on all the networks contained in the
        document, so their info can also be set.
        
        # reset the held network info
        self.networks = []
        
        # get all of the networks within the document
        network_loci = self.user.temp_locus.get_children_loci()
        for net_locus in network_loci:
            dl_net = DlNetwork(self.user, net_locus.get_type(), None, 
              net_locus, 1)
            # call the user interface to add the network
            new_view = self.uil_document_view.showNetwork(dl_net._this())
            dl_net.set_view_callback(new_view)
            
            dl_net.load_info(is_document)

            self.networks.append(dl_net)
    """
                
class DlLocus(modules.piper__POA.UIL2DL.Locus):
    
    def __init__(self, user, type, dl_parent_locus = None, 
      this_locus = None, locus_name = None):
        self.user = user 
        # if we already have a locus, we don't need to create a new one
        if this_locus:
            self.locus = this_locus
        else:
            if dl_parent_locus:
                self.locus = dl_parent_locus.locus.create_locus(type, locus_name)
            else:
                self.locus = self.user.temp_locus.create_locus(type, locus_name)
                
        self.dl_parent_locus = dl_parent_locus
        
        # Create connectors for each specified
        # input and output of the created locus.
        input_info_list = self.locus.get_connector_info('Input')
        self.local_inputs = []
        
        # create a connector for each input
        for info in input_info_list:
            new_connector = DlConnector('Input', info[0], info[1], info[2],
              self)
            self.local_inputs.append(new_connector)
        
        output_info_list = self.locus.get_connector_info('Output')
        self.local_outputs = []
        # create a connector for each output
        for info in output_info_list:
            new_connector = DlConnector('Output', info[0], info[1], info[2], 
              self)
            self.local_outputs.append(new_connector)
            
        # Create parameters for each parameter in the created locus.
        param_info_list = self.locus.get_parameters()
        self.local_parameters = []
        for info in param_info_list:
            new_param = DlParameter(info[0], info[1], info[2], info[3], 
              self.locus)
            self.local_parameters.append(new_param)
            
        # print 'self.local_inputs', self.local_inputs
            
    # --- corba functions
    def _get_name(self):
        return self.locus.get_net_name()
    def _get_type(self):
        return self.locus.get_type()
    
    def _get_inputs(self):
        inputs_to_return = []
        for n in self.local_inputs:
            inputs_to_return.append(n._this())
            
        return inputs_to_return
        
    def _get_outputs(self):
        outputs_to_return = []
        for n in self.local_outputs:
            outputs_to_return.append(n._this())
        return outputs_to_return
        
    def _get_parameters(self):
        parameters_to_return = []
        for n in self.local_parameters:
            parameters_to_return.append(n._this())
        return parameters_to_return
        
    def addInput(self, name, type, description):
        """Add an input connector to the current locus.
        
        Raises:
            * piper.UIL2DL.NameError - If an input with the given name
            already exists (the name is a unique identifier).
        """
        # print 'adding input:', name, 'to file:', self.locus.location

        try:
            return addInput_internal(name, type, description)
        except KeyError:
            raise modules.piper.UIL2DL.NameError \
              ("An input with the name %s already exists." % name)
            
    def addInput_internal(self, name, type, description):
        """Internal addInput so it can be called from non-corba as well.
        
        Raises: 
            KeyError - If there is a problem adding the connector.
        """
        text_type = conn_type_corba_to_string(type)   
        
        self.locus.add_connector('Input', name, text_type, description)
            
        # create a dl connector object
        new_connector = DlConnector('Input', name, text_type, 
          'No Description Available', self)
        self.local_inputs.append(new_connector)
        
        # call the user interface to show the new connector
        self.uil_locus_view.showNewInput(new_connector._this())
        
        return new_connector._this()
        
    def addOutput(self, name, type, description):
        """Add an output connector to the current locus.
        
        Raises:
            * piper.UIL2DL.NameError - If an output with the given name
            already exists (the name is a unique identifier).
        """               
        text_type = conn_type_corba_to_string(type)
        try:
            return addOutput_internal(name, type, description)
        except KeyError:
            raise modules.piper.UIL2DL.NameError \
              ("An output with the name %s already exists." % name)
              
    def addOutput_internal(self, name, type, description):
        """Internal addOutput to allow this to be called without CORBA and
        still deal with errors properly.
        
        Raises: 
            KeyError - If there is a problem adding the connector.
        """
        text_type = conn_type_corba_to_string(type)

        # add the connector to the xml file       
        self.locus.add_connector('Output', name, text_type, description)
      
        # create a dl connector object
        new_connector = DlConnector('Output', name, text_type, 
          'No Description Available', self)
        self.local_outputs.append(new_connector)
        
        # call the user interface to show the new connector
        self.uil_locus_view.showNewOutput(new_connector._this())
        
        return new_connector._this()
        
        
    def addParameter(self, name, type, description):
        """Add a parameter to the current locus.
        
        Raises:
            * piper.UIL2DL.NameError - If a parameter with the given name
            already exists (the name is a unique identifier).
        """
        text_type = param_type_corba_to_string(type)
        
        try:
            self.locus.add_parameter(name, text_type, description)
        except KeyError:
            raise modules.piper.UIL2DL.NameError \
              ("A parameter with the name %s already exists." % name)
              
        # create a parameter object
        new_parameter = DlParameter(name, type, 'none', 'not available', self)
        self.local_parameters.append(new_parameter)
        self.parameter.append(new_parameter._this())
        
        # call the user interface to show the new connector
        self.uil_locus_view.showNewParameter(new_parameter._this())
        
        return new_parameter._this()
        
    def removeInput(self, input_to_remove):
        """Remove an input from the current locus.
        
        Raises:
            piper.UIL2DL.LookupError - If the input cannot be found.
        """
        name_to_remove = input_to_remove._get_name()
        
        # remove the connector
        try:
            self.locus.remove_connector('Input', name_to_remove)
        except KeyError:
            raise modules.piper.UIL2DL.LookupError \
              ("The passed input connector was not found in the locus.")
              
        # find the local input that corresponds the passed input and
        # remove it
        for l_input in self.local_inputs:
            if input_to_remove._is_equivalent(l_input._this()):
                local_input_to_remove = l_input
                
        self.local_inputs.remove(local_input_to_remove)
              
        # call the user interface to remove this locus
        self.uil_locus_view.hideInput(input_to_remove)
        
    def removeOutput(self, output_to_remove, propagate_change = 1):
        """Remove an output connector from the current locus.
        
        Raises:
            piper.UIL2DL.LookupError - If the output cannot be found.
        """
        name_to_remove = output_to_remove._get_name()
        
        # remove the connector
        try:
            self.locus.remove_connector('Output', name_to_remove)
        except KeyError:
            raise modules.piper.UIL2DL.LookupError \
              ("The passed output connector was not found in the locus.")
              
        # find the local output that corresponds the passed output and
        # remove it
        for output in self.local_outputs:
            if output_to_remove._is_equivalent(output._this()):
                local_output_to_remove = output
                
        self.local_outputs.remove(local_output_to_remove)
        
        if propagate_change:
            # remove the output corresponding to this connector 
            # in the parent locus
            self.dl_parent_locus._remove_output_connector(self, 
              local_output_to_remove)
              
        # call the user interface to remove this locus
        self.uil_locus_view.hideOutput(output_to_remove)
        
    def removeParameter(self, param_to_remove):
        """Remove a parameter from the current locus.
        
        Raises:
            piper.UIL2DL.LookupError - If the parameter cannot be found.
        """
        name_to_remove = param_to_remove._get_name()
        
        # remove the parameter
        try:
            self.locus.remove_parameter(name_to_remove)
        except KeyError:
            raise modules.piper.UIL2DL.LookupError \
              ("The parameter to remove was not found in the locus.")
        
        # find the local parameter that corresponds to the parameter
        for param in self.local_parameters:
            if param_to_remove._is_equivalent(param._this()):
                local_param_to_remove = param
                
        self.local_parameter.remove(local_param_to_remove)
        
        # call the user interface to get rid of this parameter
        self.uil_locus_view.hideParameter(param_to_remove)
        
    def setPosition(self, x, y):
        self.locus.set_position(x, y)
        
    def getXPosition(self):
        x_pos, y_pos = self.locus.get_position()
        return x_pos
        
    def getYPosition(self):
        x_pos, y_pos = self.locus.get_position()
        return y_pos
        
    def save(self, save_module, save_name):
        pass
        
    def getDescription(self):
        """Return the description of this object, or 'Not Available' """
        try:
            description = self.locus.get_description()
        # if no description is set for the locus
        except KeyError:
            description = "Info for locus not available."
            
        return description
        
    # --- local functions
    def set_view_callback(self, uil_locus_view):
        """Set the view function to call when updating locus info."""
        self.uil_locus_view = uil_locus_view

                
class DlNetwork(modules.piper__POA.UIL2DL.Network, DlLocus):
    
    def __init__(self, user, type, dl_parent_locus = None, 
      this_locus = None, is_main_network = 0):
        if is_main_network:
            locus_name = 'MAIN'
        else:
            locus_name = None   
        DlLocus.__init__(self, user, type, 
         dl_parent_locus, this_locus, locus_name)
        self.user = user
        
        # the loci held in this workspace
        self.held_loci = []
        self.connections = []
        
        # flag to know if this is the main locus
        self.is_main_network = is_main_network
        
    # --- corba functions
    def _get_loci(self):
        """Return all of the loci held by the network."""
        loci_ob_list = []
        for locus in self.held_loci:
            loci_ob_list.append(locus._this())
            
        return loci_ob_list
        
    def _get_connections(self):
        """Return all of the connections in the network."""
        conn_ob_list = []
        for conn in self.connections:
            conn_ob_list.append(modules.piper.UIL2DL.Connection
              (conn[0]._this(), conn[1]._this()))
              
        return conn_ob_list
              
    def addLocus(self, type):
        try:
            new_locus = DlLocus(self.user, type, self)
            self.held_loci.append(new_locus)

            # call the user interface to add the locus
            new_view = self.uil_network_view.showLocus(new_locus._this())
            new_locus.set_view_callback(new_view)
                
        except KeyError:
            raise modules.piper.UIL2DL.TypeError \
              ("Invalid type %s passed for creating a Locus." % type)
            
    def addNetwork(self, type):
        valid_networks = ['composite', 'iterator']
        if type not in valid_networks:
            raise modules.piper.UIL2DL.TypeError \
              ("Invalid type %s passed for creating a Network." % type)     
        
        new_net = DlNetwork(self.user, type, self)
        self.held_loci.append(new_net)
        
        # call the user interface to add a network
        new_view = self.uil_network_view.showNetwork(new_net._this())
        new_net.set_view_callback(new_view)
            
    def removeLocus(self, locus):
        locus_to_remove = None
        for local_locus in self.held_loci:
            if local_locus._this()._is_equivalent(locus):
                locus_to_remove = local_locus
                
        if locus_to_remove is None:
            raise modules.piper.UIL2DL.LookupError \
              ("Locus to remove was not found in the network.")
              
        self.locus.remove_locus(locus_to_remove.locus)
        
        # call the user interface to remove this locus
        self.uil_network_view.hideLocus(locus)
                
    def connectLoci(self, in_locus, in_connector, out_locus, out_connector):
        # get the local loci and connectors
        (in_local_locus, in_local_connector, out_local_locus,
          out_local_connector) = self._get_local_objects(in_locus, in_connector,
          out_locus, out_connector)
        
        # check for connector types that don't match
        if ((in_local_connector.type != out_local_connector.type) and
          (in_local_connector.type != modules.piper.UIL2DL.Connector.P_ANY and
          out_local_connector.type != modules.piper.UIL2DL.Connector.P_ANY)):
            raise modules.piper.UIL2DL.TypeError \
              ("The input type %s does not match the output type %s" % 
                (in_local_connector.type, out_local_connector.type))
    
        # make the two connections in the individual loci
        self.locus.connect_loci(in_local_locus.locus, in_local_connector.name,
          out_local_locus.locus, out_local_connector.name)
          
        #in_local_locus.add_connection('Input', in_local_connector.name, 
        #  out_local_locus.locus.id, out_local_connector.name)
        #out_local_locus.add_connection('Output', out_local_connector.name, 
        #  in_local_locus.locus.id, in_local_connector.name)
          
        # add the connection to the list of connections in the network
        self.connections.append((in_local_connector, out_local_connector))
        
        # call the user interface to make the connection
        self.uil_network_view.showConnection(in_locus, in_connector, out_locus,
          out_connector)
          
        
    def disconnectLoci(self, in_locus, in_connector, out_locus, 
      out_connector):
        # get the local loci and connectors
        (in_local_locus, in_local_connector, out_local_locus,
          out_local_connector) = self._get_local_objects(in_locus, in_connector,
          out_locus, out_connector)
          
        self.locus.disconnect_loci(in_local_locus.locus, in_local_connector.name,
          out_local_locus.locus, out_local_connector.name)
          
        # remove the connection from the list of connections in the network
        self.connections.remove((in_local_connector, out_local_connector))
        
        # call the user interface to disconnect the loci
        self.uil_network_view.hideConnection(in_locus, in_connector, out_locus,
          out_connector)
        
    def _get_local_objects(self, in_locus, in_connector, out_locus, 
      out_connector):
        """Return local connectors and loci corresponding to omniORB types."""
        # loci
        in__local_locus = None
        out_local_locus = None
        for local_locus in self.held_loci:
            if local_locus._this()._is_equivalent(in_locus):
                in_local_locus = local_locus
            elif local_locus._this()._is_equivalent(out_locus):
                out_local_locus = local_locus
                
        # get the local connectors for each passed omniORB type
        in_local_connector = None
        for input in in_local_locus.local_inputs:
            if input._this()._is_equivalent(in_connector):
                in_local_connector = input
        
        out_local_connector = None
        for output in out_local_locus.local_outputs:
            if output._this()._is_equivalent(out_connector):
                out_local_connector = output
                
        return (in_local_locus, in_local_connector, out_local_locus, 
          out_local_connector)
        
    # --- local functions
    def add_connection(self, output_locus, out_conn_name, input_locus, 
      in_conn_name):
        """Add a connection based on loading information."""
        # first find the dl loci for the input and output
        for dl_locus in self.held_loci:
            if output_locus.id_name == dl_locus.locus.id_name:
                dl_output_locus = dl_locus
            elif input_locus.id_name == dl_locus.locus.id_name:
                dl_input_locus = dl_locus
                
        # now find the connectors in each locus that contribute to the link
        for output in dl_output_locus.local_outputs:
            if output.name == out_conn_name:
                dl_out_connector = output
                
        for input in dl_input_locus.local_inputs:
            if input.name == in_conn_name:
                dl_in_connector = input
                
        # add the connection to the list of connections in the network
        self.connections.append((dl_in_connector, dl_out_connector))
        
        # call the user interface to make the connection
        self.uil_network_view.showConnection(dl_input_locus._this(), 
          dl_in_connector._this(), dl_output_locus._this(),
          dl_out_connector._this())
                
        
        
    def _add_input_connector(self, input_locus, input_connector):
        """Convenience function to add an input connector to the network.
        """
        in_name = input_locus.locus.get_net_name() + '--' + \
          input_connector.name
            
        self.addInput_internal(in_name, input_connector.type, 
          input_connector.getDescription())
          
    def _add_output_connector(self, output_locus, output_connector):
        """Convenience function to add an input connector to the network.
        """
        out_name = output_locus.locus.get_net_name() + '--' + \
          output_connector.name
        self.addOutput_internal(out_name, output_connector.type, 
          output_connector.getDescription())
          
    def _remove_input_connector(self, input_locus, input_connector):
        """Convenience function to remove an input connector from the network.
        """
        in_name = input_locus.locus.get_net_name() + '--' + \
          input_connector.name
        # now find the input corresponding to this name   
        for l_input in self.local_inputs:
            if l_input.name == in_name:
                local_input_to_remove = l_input
                
        self.removeInput(local_input_to_remove._this())
        
    def _remove_output_connector(self, output_locus, output_connector):
        """Convenience function to remove an output connector from the network.
        """
        out_name = output_locus.locus.get_net_name() + '--' + \
          output_connector.name
        # find the output corresponding to this name
        for l_output in self.local_outputs:
            if l_output.name == out_name:
                local_output_to_remove = l_output
        self.removeOutput(local_output_to_remove._this())
        
    def set_view_callback(self, uil_network_view):
        """Set the user interface corba object to call when updating the network
        """
        DlLocus.set_view_callback(self, uil_network_view)
        self.uil_network_view = uil_network_view
        
    def add_initialized_network(self, type, init_locus):
        """Add a network to this network from an already initialized locus.
        
        Returns:
            The created DlNetwork.
        """
        new_dl_net = DlNetwork(self.user, type, self, init_locus)
        new_view = self.uil_network_view.showNetwork(new_dl_net._this())
        new_dl_net.set_view_callback(new_view)
        self.held_loci.append(new_dl_net)
                
        return new_dl_net
        
    def add_initialized_locus(self, type, init_locus):
        """Add a DlLocus to this network from an already initialized locus.
        
        Returns:
            The created DlLocus.
        """
        new_dl_locus = DlLocus(self.user, type, self, init_locus)
        new_view = self.uil_network_view.showLocus(new_dl_locus._this())
        new_dl_locus.set_view_callback(new_view)
        self.held_loci.append(new_dl_locus)
        
        return new_dl_locus

    """
    def load_info(self, is_document = 0):
        Load all the contained loci and connections for this network.
        
        self.held_loci = []
        self.connections = []
        
        # get all of the loci within this network
        all_loci = self.locus.get_children_loci()
        for this_locus in all_loci:
            if (this_locus.get_type() == 'composite' or 
              this_locus.get_type() == 'iterator'):
                dl_locus = DlNetwork(self.user, this_locus.get_type(), 
                  self.locus, self, this_locus)
                
                # call the user interface to add the network
                new_view = self.uil_network_view.showNetwork(dl_locus._this())
                dl_locus.set_view_callback(new_view)
                
                dl_locus.load_info(is_document)
            else:
                dl_locus = DlLocus(self.user, this_locus.get_type(), 
                  self, this_locus)
                
                # call the user interface to add the locus
                new_view = self.uil_network_view.showLocus(dl_locus._this())
                dl_locus.set_view_callback(new_view)
                
            self.held_loci.append(dl_locus)
            
            # add new connectors to the network to represent the 
            # unconnected connectors of the new locus
            for new_input in dl_locus.local_inputs:
                if is_document:
                    #self.uil_network_view.showNewInput(new_input._this())
                    # self._add_input_connector(dl_locus, new_input)
                    in_name = dl_locus.locus.get_net_name() + '--' + \
                      new_input.name
            
                    self.addInput_internal(in_name, new_input.type, 
                      new_input.getDescription(), 0)
                else:
                    self.uil_network_view.showNewInput(new_input._this())
                
            for new_output in dl_locus.local_outputs:
                if is_document:
                    #self.uil_network_view.showNewOutput(new_output._this())
                    #self._add_output_connector(dl_locus, new_output)
                    out_name = dl_locus.locus.get_net_name() + '--' + \
                      new_output.name
            
                    self.addOutput_internal(out_name, new_output.type, 
                      new_output.getDescription(), 0)
                else:
                    self.uil_network_view.showNewOutput(new_output._this())
            
        # get all of the connections within this network
        # This sucks, I need to think of a better way to do it...
        
        for dl_locus in self.held_loci:
            # print 'type:', dl_locus.locus.get_type()
            conn_inputs = dl_locus.locus.get_connected_links('Input')
            # print 'input_info', conn_inputs
            for input_info in conn_inputs:
                # get the dl_locus that is the output of the link
                for in_dl_locus in self.held_loci:
                    if in_dl_locus.locus == input_info[1]:
                        output_locus = in_dl_locus
                        
                # now we have the input and output, and need to get the
                # connectors and add them to the connections in this locus
                #  'local_inputs', dl_locus.local_inputs
                for input in dl_locus.local_inputs:
                    input_connector = input.get_connection()
                    if input_connector:
                        incoming_locus = input_connector.getLocus()
                        if incoming_locus._is_equivalent(output_locus._this()):
                            self.connections.append(
                              (input, input.get_connection()))
                              
                            # call the user interface to add the connection
                            self.uil_network_view.showConnection \
                              (input.getLocus(), input._this(), 
                               output_locus._this(), 
                               input.get_connection()._this())
                               
                            # remove the connectors for the network representing
                            # the ones involved in the new connection
                            self.uil_network_view.hideInput(input._this())
                            output = input.get_connection()
                            self.uil_network_view.hideOutput(output._this())
    """
                        
class DlConnector(modules.piper__POA.UIL2DL.Connector):
    
    def __init__(self, io_type, name, type, description, dl_locus):
        self.io_type = io_type
        self.parent_dl_locus = dl_locus
        self.name = name
        self.description = description
        self.type = conn_type_string_to_corba(type)
            
    # --- corba functions
            
    def _get_name(self):
        return self.name
        
    def _set_name(self, name):
        self.name = name
        # TODO: Need to propograte this change back to the XML file.
        
    def _get_type(self):
        return self.type
        
    def _set_type(self, type):
        self.type = type
        # TODO: Need to propagate this change back to the XML file.
        
    def getDescription(self):
        """Return the description of this object, or 'Not Available' """
        return self.description
        
    def getLocus(self):
        """Return the locus that contains this connector."""
        return self.parent_dl_locus._this()
        
    def mapToParent(self):
        """Map this connector to the parent composite (as a 'netConnector').

        Returns:
        * The newly created connector.
        """
        # munge together the unique network name of the locus and the
        # connector name to creat the unique name of the new netConnector
        new_conn_name = (self.parent_dl_locus.locus.get_net_name() + '::' +
                         self.name)

        # add the new connector to the composite locus containing this
        # connector
        self.parent_dl_locus.dl_parent_locus.locus.add_connector(
            self.io_type, self.name,  new_conn_name, self.type, self.description,
            self.parent_dl_locus.locus)

        # get a new connector for the parent locus and return this
        new_conn_type = conn_type_string_to_corba(self.type)
        if self.io_type == 'Input':
            return self.parent_dl_locus.addInput(new_conn_name, new_conn_type,
                                                 self.description)
        else:
            return self.parent_dl_locus.addOutput(new_conn_name, new_conn_type,
                                                  self.description)

    # --- local functions
        
    def get_connection(self):
        """Return the connector this locus is connected to.
        
        Returns None if it is not connected.
        """
        other_locus = None
        locus_links = self.parent_dl_locus.locus.get_connected_links\
                      (self.io_type)
        # print 'locus_links:', locus_links
        # print 'self.name:', self.name
        for link_info in locus_links:
            if link_info[0] == self.name:
                other_locus = link_info[1]
                other_conn_name = link_info[2]
        
        if other_locus is None:
            return None
        
        # now get the other possible loci this could be linked to
        our_network = self.parent_dl_locus.dl_parent_locus
        for dl_locus in our_network.held_loci:
            # find the locus that matches the one we link to
            if dl_locus.locus == other_locus:
                # get the right connector to link with
                if self.io_type == 'Input':
                    connectors = dl_locus.local_outputs
                if self.io_type == 'Output':
                    connectors = dl_locus.local_inputs
                    
                for connector in connectors:
                    if connector.name == other_conn_name:
                        match_connector = connector
                        
        return match_connector
        
class DlParameter(modules.piper__POA.UIL2DL.Parameter):
    
    def __init__(self, name, type, value = 'none', description = '', 
      locus = None):
        self.parent_locus = locus
        self.name = name
        self.type = param_type_string_to_corba(type)
        self.description = description
        if value == 'none':
            self.value = None
        else:
            self.value = value
        
    def _get_name(self):
        return self.name
        
    def _set_name(self, name):
        self.name = name
        # TODO: Need to propagate this change back to the XML file.
        
    def _get_type(self):
        return self.type
        
    def _set_type(self, type):
        self.type = type
        # make the change in the XML file
        type_to_set = param_type_corba_to_string(self.type)
        self.parent_locus.set_parameter_type(self.name, type_to_set)
        
    def setInfo(self, info_to_set):
        """Set the value parameter with the specified information.
        
        This expects the information to be in the same format as the 
        specified type, and tries to check this. If there are any problems,
        then we assume the passed info is bad, and raise a TypeError.
        """
        self.value = info_to_set.value()
        
        # make the change in the XML file
        self.parent_locus.set_parameter_info(self.name, self.value)
    
    def getInfo(self):
        """Return the information associated with the parameter.
        """     
        if self.value is None:
            raise modules.piper.UIL2DL.LookupError \
              ("There is no information available for this parameter.")
              
        if self.type == modules.piper.UIL2DL.Parameter.P_STRING:
            type_code = CORBA._tc_string
        # default to returning a string as well 
        else:
            type_code = CORBA._tc_string
              
        return CORBA.Any(type_code, self.value)
        
        
    def getDescription(self):
        """Return the description of this object, or 'Not Available' """
        return self.description
        
class DlProcess(modules.piper__POA.UIL2DL.Process):
    """Represent a process that has been submitted to the bl."""
    
    def __init__(self, bl_connection):
        """Initialize with the connection to the brokering layer."""
        self.bl_connection = bl_connection
        
    # --- corba functions
    def queryStatus(self):
        """Query the overall status of the process.
        
        Returns a Status struct. The first element is a string describing
        the status of the process and the second is a double indicating
        the progress of the process towards completion (1 = complete;
        0 = not started).
        """
        string_info, double_info = bl_connection.query_status()
        
    def queryLocusStatus(self, locus):
        """Query the status of a particular locus."""
        pass
    
        
        
        
        
