#!/usr/bin/env python
"""addpiperpass.py

A really simple user interface script for adding passwords to the dictionary
for a user.
The idea is that passwords for use in piper are stored in a password 
database protected by a single username and password. Thus, each individual 
user can store all of the passwords for use in piper together and access 
them with one global piper password that will be submitted when the 
front/user interface logs into the definition layer.
The purpose of this script is then to set the access passwords before 
running piper, and then piper will be able to access them during running
using the user interface login id and password.
Right now I'm using this only for storing database access passwords, but
the plan is to expand it to include passwords for communication with the 
brokering layer.

Use:
	./addpiperpass.py <options> 
	
	Options
	--------
	    Required
	    --------
		    -u username : The username to enter the password under.
		    -t type_of_password : The type of password to add. Accepted types:
			    o 'database'
		Optional
		--------
		    -n : New - Create a new user name and password database for the
			username.
			
"""
# standard modules
import sys
import getopt
import getpass

# piper passwords stuff
from modules.piperpasswords import PiperPass

### constants
# usage message
USAGE_MSG = 'Usage: piperpass <options>\n' + \
            'Where <options> include:\n' + \
			'-u <username> (required) The username to access the password database\n' + \
			'-t <type>     (required) The type of password to add.\n' + \
			'                         Valid options for <type> are: \n' + \
			'                           o database - Add a database password \n' + \
			'-n            (optional) Create a new username/password. \n'
      
def main(argv):
	(options, arguments) = getopt.getopt(sys.argv[1:], 'u:t:n')
	
	# print a usage message if the right number of arguments aren't passed
	if ((len(options) > 3) or (len(options) < 2) or (len(arguments) > 0)):
		sys.stderr.write(USAGE_MSG)
		
		return 1
		
	# get the username and type to add
	username = None
	add_type = None
	for n in options:
		if (n[0] == '-u'):
			username = n[1]
		elif (n[0] == '-t'):
			add_type = n[1]
			
	# raise errors if we don't have a username and password.
	if ((username == None) or (add_type == None)):
		sys.stderr.write(USAGE_MSG)
		
		return 1
	
	# determine if we are adding a new user/password
	new_user = 0
	for n in options:
		if n[0] == '-n':
			new_user = 1
		
	if new_user:
		print 'Adding the new user name to the password database...'
		# get the password
		user_pass = getpass.getpass('Enter a password to access this username: ')
		second_pass = getpass.getpass('Please confirm this password: ')
		
		if user_pass != second_pass:
			sys.stderr.write('Sorry. The entered passwords do not match. \n')
			sys.stderr.write('Please try again :-) \n')
			return 1
		else:
			pass_db = PiperPass()
			pass_db.add_new_user(username, user_pass)
			print 'New username and password successfully added to password database'
			
	else:
		# get the password
		user_pass = getpass.getpass('Enter a password to access this username: ')
		try:
			pass_db = PiperPass(username, user_pass)
			print 'Successfully logged into password database.'
		except KeyError:
			pass_db.close()
			sys.stderr.write('Sorry. That username and password are not valid. \n')
			sys.stderr.write('Use the -n option to add a new username. \n')
			
			return 1
			
	# add the appropriate password type to the database
	if (add_type == 'database'):
		add_db_password(pass_db)
	else:
		sys.stderr.write \
		  ('Cannot create a password entry for the specified type. \n')
		sys.stderr.write(USAGE_MSG)
def add_db_password(pass_db):
	"""Add a database password to an opened password database.
	
	Arguments:
		pass_db - An opened password database to add the new password
		information to.
	"""
	print 'Adding a password for database connection...'
	db_specific_info = {}
	# get all of the information we need to add to add
	db_specific_info['database'] = raw_input \
	  ('Enter the database type (ie. mysql) for connecting to: ')
	  
	db_specific_info['username'] = raw_input \
	  ('Enter the username to connect to the database with: ')
	  
	pword = getpass.getpass('Enter the password to connect with: ')
	
	pass_db.add_password('database', pword, db_specific_info)
	print 'The password was successfully added.'
	pass_db.close()
	
if __name__ == '__main__':
	sys.exit(main(sys.argv))