from gtk import *
from gnome.ui import *

from piperconfig import *

import string

class fileDialog(GtkFileSelection):
	"""
	This class provides a modal Gtk GUI for selecting
	a file. This is based on the _FileSelection class by James Henstridge
	in the GtkExtras of the pyGtk distribution.
	"""
	def __init__(self):
		GtkFileSelection.__init__(self)
		self.connect("destroy", self.quit)
		self.connect("delete_event", self.quit)
		self.ok_button.connect("clicked", self.setFile)
		self.cancel_button.connect("clicked", self.quit)
		self.hide_fileop_buttons()
		self.selectedFile = None
		# make modal
		grab_add(self)

	def quit(self, *args):
		"""
		This cleans everything up when the window is exited. It gets
		rid of the window and exits the mainloop started in the call
		to the selection window.
		"""
		self.hide()
		self.destroy()
		mainquit()
	
	def setFile(self, button):
		"""
		This function is called by the Gtk file selection widget when a file
		is selected. It sets the currently selected file as this file.
		"""
		self.selectedFile = self.get_filename()
		self.quit()
		
class dbSelectDialog(GtkDialog):
	"""
	Provides a dialog for connecting to a database using Loci.
	The selection dialog will allow the user to enter the type of
	database (ie. MySQL, PostgreSQL, Sybase, etc...), the database
	name to connect to, and hostname, user and password stuff. 
	
	Note: This dialog assumes it was created in a mainloop(), so it calls 
	mainquit() upon exiting.
	"""
	
	def __init__(self):
		"""
		This extends the basic Gtk Dialog class. The initialization
		function does everything for you--creates the dialog, fills 
		the selection boxes with information from XML files (if available)
		and gets the users choices. 
		
		After the user makes their choice, the following class variables
		are available:
			
			- DBtype: The database type selected (MySQL, Sybase, etc).
			- host: The host to connect to.
			- user: The user to connect as.
			- password: The password to use. 
			  ***Note: The password returned is encrypted to prevent
			  password sniffing.
			- port: TCP/IP port to connect to (if not set, will be the
			  default MySQL port)
			- socket: location of unix_socket to use (if not set, will
			  use TCP)
			
		Note: Any variables not set during the dialog will returned as ''.
		"""
		GtkDialog.__init__(self)
		
		# set up the database parameters to get
		self.DBtype = ''
		self.host = ''
		self.user = ''
		self.password = ''
		self.port = ''
		self.socket = ''
	
		# set up the dialog
		self.set_title('Select a Database to Connect to')
		self.set_uposition(dialog_upos_x, dialog_upos_y)
		
		table = GtkTable(8, 4, FALSE)
		table.set_row_spacings(25)
		table.set_col_spacings(5)
		table.show()
		self.vbox.add(table)
		
		self.typeChoice = GnomeEntry('type')
		self.typeChoice.show()
		self.typeChoice.list.connect('selection_changed', self.updateDBtype)
		typeLabel = GtkLabel('Database Type')
		typeLabel.show()
		table.attach(typeLabel, 0, 1, 1, 2)
		table.attach(self.typeChoice, 1, 2, 1, 2)
		
		self.hostChoice = GnomeEntry('host')
		self.hostChoice.show()
		hostLabel = GtkLabel('Host')
		hostLabel.show()
		table.attach(hostLabel, 0, 1, 2, 3)
		table.attach(self.hostChoice, 1, 2, 2, 3)
		
		self.userChoice = GnomeEntry('user')
		self.userChoice.show()
		userLabel = GtkLabel('User')
		userLabel.show()
		table.attach(userLabel, 0, 1, 3, 4)
		table.attach(self.userChoice, 1, 2, 3, 4)
		
		self.passChoice = GtkEntry()
		self.passChoice.set_editable(TRUE)
		self.passChoice.set_visibility(FALSE)
		self.passChoice.show()
		passLabel = GtkLabel('Password')
		passLabel.show()
		table.attach(passLabel, 0, 1, 4, 5)
		table.attach(self.passChoice, 1, 2, 4, 5)
		
		self.portChoice = GnomeEntry('port')
		self.portChoice.show()
		portLabel = GtkLabel('Port')
		portLabel.show()
		table.attach(portLabel, 0, 1, 5, 6)
		table.attach(self.portChoice, 1, 2, 5, 6)
		
		self.sockChoice = GnomeEntry('sock')
		self.sockChoice.show()
		sockLabel = GtkLabel('Socket')
		sockLabel.show()
		table.attach(sockLabel, 0, 1, 6, 7)
		table.attach(self.sockChoice, 1, 2, 6, 7)
		
		ok_button = GtkButton('OK')
		ok_button.show()
		cancel_button = GtkButton('Cancel')
		cancel_button.show()
	
		self.action_area.pack_start(ok_button)
		self.action_area.pack_start(cancel_button)
		
		self.connect("destroy", self.quit)
		self.connect("delete_event", self.quit)
		ok_button.connect("clicked", self.setDBinfo)
		cancel_button.connect("clicked", self.quit)
		# make modal
		grab_add(self)
		
		self.show()
		
	def updateDBtype(self, button):
		"""
		This is used to update the pull down boxes of choices when a 
		new database choice (ie. choose PostgreSQL instead of MySQL)
		is selected from the Database Type menu.
		
		Not yet implemented
		"""
		pass
		
	def setDBinfo(self, *args):
		"""
		Sets all the appropriate internal variables when the OK button
		is clicked and cleans up the window.
		"""
		# get all of the information from the various boxes
		# and clean up whitespace issues
		typeInfo = string.strip(self.typeChoice.entry.get_text())
		hostInfo = string.strip(self.hostChoice.entry.get_text())
		userInfo = string.strip(self.userChoice.entry.get_text())
		passInfo = string.strip(self.passChoice.get_text())
		portInfo = string.strip(self.portChoice.entry.get_text())
		sockInfo = string.strip(self.sockChoice.entry.get_text())
		

		self.DBtype = typeInfo
		self.host = hostInfo
		self.user = userInfo
		self.password = passInfo
		self.port = portInfo
		self.socket = sockInfo
			
		self.quit()
		
	def quit(self, *args):
		"""
		This cleans everything up when the window is exited. It gets
		rid of the window and exits the mainloop started in the call
		to the selection window.
		"""
		self.hide()
		self.destroy()
		mainquit()