# standard modules
import sys
import os

# Gtk/Gnome stuff
from gtk import *
from gnome.ui import *
from GdkImlib import *

# import CORBA stubs
import modules.piper.UIL2DL

# windowlet
from windowlet import PiedWindowlet

# terminals
from terminal import PiedTerminal

# configuration variables
from piperconfig import *

# --- constants
SYMBOL_DIR = os.path.join(piper_pixmap_dir, 'symbols')
ICON_DIR = os.path.join(piper_pixmap_dir, 'icons')


class PiedNode:
	""" 
	This class describes the information about a particular node
	and provides functions allowing manipulation of node attributes.
	"""
	def __init__(self, network, type = 'network', initialx = 0, 
	  initialy = 0, dl_node = None):
		""" 
		Initializes a node. The network containing the node must be 
		passed as a parameter. The node is also passed the initial x 
		and y position where the node is to be drawn, and the type of nodes
		to be created. If this node is the background/network node, the 
		node should be initialized with initial parameters of ('network', 0,0)
		for the type and position. In this case, a separate GUI will not need to 
		be initialized.
		"""

		# add network's object to node
		self.network = network

		# create the definition layer node to query for information 
		# about the node
		
		# for subnet nodes we need to add a new network
		if type == 'subnet' or type == 'iterator':
			# if we are passed a dl_node, that is the network and node
			if dl_node:
				self.network = dl_node
				self.dl_node = dl_node
			# otherwise, create one
			else:
				self.network = network.network.addNetwork()
				self.network._narrow(modules.piper.UIL2DL.Network)		
				self.dl_node = self.network
		# need to ignore this so we don't create too many network/composites.
		# this gets called twice -- once for add_node, and then again in the 
		# initialization of the network--don't do anything the second time.
		elif type == 'network':
			pass
		else:
			# if we are passed a dl_node, use that as our node
			if dl_node:
				self.dl_node = dl_node
			# otherwise create one
			else:
				self.dl_node = network.network.addnode(type)
				self.dl_node._narrow(modules.piper.UIL2DL.node)
		
		# initialize node's input and output lists
		self.input = []
		self.output = []

		self.type = type
		self.name = ''
		self.label = type
# ??? temporary
		self.message = 'Short message'
#		self.message = 'This is a very long message.  This point is to\n' + \
#						'to see how this fits in the message box and in\n' + \
#						'the tooltips window.'
		self.initial_x = initialx
		self.initial_y = initialy
		self.world_x = initialx
		self.world_y = initialy
		
		# if we have a non-network/background nodes, we need to initialize
		# the parameters for the GUI. Otherwise, don't bother with it.
		if ((initialx != 0) and (initialy != 0) and (self.type != 'network')):
			# set the position of the node in the dl
			self.dl_node.setPosition(initialx, initialy)
			
			# query the dl for information we will need about the node
			self.number_inputs_total = len(self.dl_node._get_inputs())
			self.number_outputs_total = len(self.dl_node._get_outputs())
		
			# for the GUI name, the idea is that if the node has a 
			# custom GUI then it will have a parameter with the name "widget"
			# So look through the parameter for this, and if we don't find it
			# assume a "default" gui_name.
			self.gui_name = 'default'
			for param in self.dl_node._get_parameters():
				if param._get_name() == 'widget':
					self.gui_name = param.getInfo().value()

		# set up a group of widgets in the network to represent the node
		self.group = self.network.root().add('group')

		# information about the dots of the node
		self.dot_text = ' '

		# the default theme and name for the symbol
		self.symbol_theme = 'simple' # temporary
		self.symbol_name = type

		# the default parameters for the symbol
		self.icon_theme = 'cute'
		self.icon_name = 'people'

		# the default user preference color (blue)
		self.color_selected = self.network.get_color('blue')

		# how many inputs and outputs for a node are connected
		self.number_inputs_connected = 0
		self.number_outputs_connected = 0

# ??? Is this needed here?
			# the components of the box holding the widget
#			self.windowlet.ebox = GtkEventBox()
#			self.windowlet.frame = GtkFrame()
#			self.windowlet.pane = GtkFrame()
#			self.scrolled_window = GtkScrolledWindow()
			
			# parameter to determine if the GUI window is shown
#			self.windowlet.shown = 1			
			
	def buildNodeGUI(self):
		"""
		This creates the initial GUI for a node. This should be called
		for nodes in a network that need a GUI (and not for "network" nodes).
		"""
		#### set up the colors/fonts etc from the network paraemeters
		self.color_white = self.network.color_white
		self.color_black = self.network.color_black
		self.color_red = self.network.color_red
		self.color_yellow = self.network.color_yellow
		self.color_green = self.network.color_green
		self.color_blue = self.network.color_blue
		self.wscolor_bg = self.network.wscolor_bg
		self.wscolor_fg = self.network.wscolor_fg
# needs to be in config file
		# set default colors for the background and foreground
		self.color_bg = self.network.color_bg
		self.color_fg = self.network.color_fg

		# scale must be in whole numbers!
		scale = 1
		# self.space must be a multiple of 4 so that a quarter_space is whole!
		self.space = 4 * scale
		self.double_space = self.space * 2
		self.half_space = self.space / 2
		self.quarter_space = self.space / 4
		self.line_width = scale
		self.text_box_margin = 2 * scale
		# adobe helvetica sizes are 80 100 120 140 180 240  
		text_font_size = '120'
		number_font_size = '80'
		font_weight = 'medium'
		self.text_font = '-adobe-helvetica-' + font_weight + '-r-normal-*-*-' + text_font_size + '-*-*-*-*-*-*'
		self.number_font = '-adobe-helvetica-' + font_weight + '-r-normal-*-*-' + number_font_size + '-*-*-*-*-*-*'
		
		
		###initialize the network representation of the node
		
		# input/output indices must always match terminal_from/to_input/output indices
		# for example, input[3] is connected to terminal_from_input[3]
		# initialize list of terminal_from_input's
		self.group.terminal_from_input = []
		# initialize list of terminal_to_output's
		self.group.terminal_to_output = []

		# set dot positions based on the position of the GUI
		self.dot_x = self.world_x
		self.dot_y = self.world_y
		
		# add all of the parts to the GUI. These must be called in this specific order.
		self._addDotText()
		self._addDotPicture()
		self._addSymbol()
		self._addIcon()
		self._addText()
		self._addTextBox()
		self._addSelectionBox()
		self._add_initial_inputs()
		self._add_initial_outputs()
		self._arrangeGUI()
		self._addInitialWidget()
		
	def _addDotText(self):
		"""
		This function is called by buildnodeGUI, and adds 
		the dot text to the gui representation of the node
		"""
		# put dot text in node group
		self.group.dot_text = self.group.add('text',text=self.dot_text,  
		                                       x=self.dot_x, y=self.dot_y,  
                                               font=self.number_font,  
                                               fill_color_gdk=self.color_fg,  
                                               anchor=ANCHOR_CENTER)
		
		# set_data used to attach python object to canvas item (dot_text)
		self.group.dot_text.set_data('node', self)
		self.group.dot_text.set_data('terminal', None)
			
		# attach node text events to the network node event handlers
		self.group.dot_text.connect('event', self.network.node_event, self)
			

			
	def _addDotPicture(self):
		"""
		This function is called by buildnodeGUI, and adds the 
		dot picture to the gui representation of the node.
		"""		
		# determine dot size - based on size of text item
		number_half_width = self.group.dot_text['text_width'] / 2
		number_half_height = self.group.dot_text['text_height'] / 2
		# which is bigger, width or height?
		if number_half_width > number_half_height:
			self.dot_half_width = number_half_width + self.quarter_space
		else:
			self.dot_half_width = number_half_height + self.quarter_space

		self.group.dot = self.group.add('rect',
										x1=self.dot_x-self.dot_half_width,
										y1=self.dot_y-self.dot_half_width,
										x2=self.dot_x+self.dot_half_width,
										y2=self.dot_y+self.dot_half_width,
										width_units=self.line_width,
										fill_color_gdk=self.color_blue,
										outline_color_gdk=self.color_fg)
  
		# set_data used to attach python object to canvas item (dot)
		self.group.dot.set_data('node', self)
		self.group.dot.set_data('terminal', None)
		
		# attach events to the network event handlers	
		self.group.dot.connect('event', self.network.node_event, self)
		
	def _addSymbol(self):
		"""
		This function is called by buildnodeGUI, and adds the symbol
		picture to the GUI representation of the node.
		"""
		# set the symbol for the gui based on the theme
		
		try:
			self.symbol = Image(os.path.join(SYMBOL_DIR, self.symbol_theme, 
		      self.symbol_name + '.xpm'))
		# if we get an IOError then we don't have a special symbol
		# for this node and just use the default.
		except IOError:
			self.symbol = Image(os.path.join(SYMBOL_DIR, self.symbol_theme, 
		      'widget.xpm'))
		symbol_image = self.symbol._im
		symbol_x = self.world_x
		symbol_y = self.world_y
		self.group.symbol = self.group.add('image', image=symbol_image, x=symbol_x, y=symbol_y,  
                                           width=self.symbol.rgb_width,  
                                           height=self.symbol.rgb_height,  
                                           anchor=ANCHOR_CENTER)

		# set_data used to attach python object to canvas item (symbol)
		self.group.symbol.set_data('node', self)
		self.group.symbol.set_data('terminal', None)
		# attach events to the network event handlers
		self.group.symbol.connect('event', self.network.node_event, self)

# temp
		self.group.symbol.hide()
		self.symbol.rgb_height = 0

	def _addIcon(self):
		"""
		This function is called by buildnodeGUI, and adds the icon 
		picture to the GUI representation of the node.
		"""
		icon = Image(os.path.join(ICON_DIR, self.icon_theme, 
		  self.icon_name + '.xpm'))
		
		# make a pixmap from the icon for a dnd icon
		pix_win = GtkWindow()
		self.icon_pix, self.icon_mask = create_pixmap_from_xpm(pix_win, None,
		  os.path.join(ICON_DIR, self.icon_theme, self.icon_name + '.xpm'))
		
		icon_image = icon._im
		icon_x = self.world_x
		icon_y = self.world_y
		self.group.icon = self.group.add('image',  
                                           image=icon_image,  
                                           x=icon_x,  
                                           y=icon_y,  
                                           width=icon.rgb_width,  
                                           height=icon.rgb_height,  
                                           anchor=ANCHOR_CENTER)
										   
		# set_data used to attach python object to canvas item (icon)
		self.group.icon.set_data('node', self)
		self.group.icon.set_data('terminal', None)

		self.group.icon.connect('event', self.network.node_event, self)

# temp
		self.group.icon.hide()
		icon.rgb_height = 0

	def _addText(self):
		"""
		Called by buildnodeGUI, this adds the text description to the
		node. This is initially set as self.label, the type of the
		node.
		"""
		# text position temporarily set to node's
		self.group.text = self.group.add('text',  
                                           text=self.label,  
                                           x=self.world_x,  
                                           y=self.world_y,  
                                           font=self.text_font,  
                                           fill_color_gdk=self.color_fg,  
                                           anchor=ANCHOR_CENTER)
		
		# get node text size and reset the position
		self.text_x = self.world_x
		if self.symbol.rgb_height > self.dot_half_width:
			self.text_y = self.world_y + self.symbol.rgb_height / 2	\
								+ self.double_space	\
								+ self.group.text['text_height'] / 2
		else:
			self.text_y = self.world_y + self.dot_half_width	\
								+ self.double_space	\
								+ self.group.text['text_height'] / 2
		self.group.text.set(y=self.text_y)

		# set_data used to attach python object to canvas item (text)
		self.group.text.set_data('node', self)
		self.group.text.set_data('terminal', None)

		self.group.text.connect('event', self.network.node_event, self)
		
	def _addTextBox(self):
		"""
		This function is called by buildnodeGUI, and adds the node 
		text box to the GUI representation of the node.
		"""
		# put node text_box on Network
		self.text_box_half_width = self.group.text['text_width'] / 2 + self.text_box_margin
		self.text_box_half_height = self.group.text['text_height'] / 2 + self.text_box_margin
		text_box_x1 = self.text_x - self.text_box_half_width
		text_box_y1 = self.text_y - self.text_box_half_height
		text_box_x2 = self.text_x + self.text_box_half_width
		text_box_y2 = self.text_y + self.text_box_half_height
		
		self.group.text_box = self.group.add('rect',  
                                           x1=text_box_x1,  
                                           y1=text_box_y1,  
                                           x2=text_box_x2,  
                                           y2=text_box_y2,  
                                           width_units=self.line_width,  
                                           fill_color_gdk=self.color_bg,  
                                           outline_color_gdk=self.color_fg)

		# set_data used to attach python object to canvas item (text_box)
		self.group.text_box.set_data('node', self)
		self.group.text_box.set_data('terminal', None)

		self.group.text_box.connect('event', self.network.node_event, self)
		
	def _addSelectionBox(self):
		"""
		This function is called by buildnodeGUI, and adds the node 
		selection box to the GUI representation of the node.
		"""
		# get the square bounds of the node (in item-relative coordinates)
		node_bounds = self.group.get_bounds()

		selection_box_x1 = node_bounds[0] - self.space
		selection_box_y1 = node_bounds[1] - self.space
		selection_box_x2 = node_bounds[2] + self.space
		selection_box_y2 = node_bounds[3] + self.space
		
		# put node selection_box on Network
		self.group.selection_box = self.group.add('rect',  
                                           x1=selection_box_x1,  
                                           y1=selection_box_y1,  
                                           x2=selection_box_x2,  
                                           y2=selection_box_y2,  
                                           width_units=self.line_width,  
                                           fill_color_gdk=None,  
                                           outline_color_gdk=self.color_selected)

		# set_data used to attach python object to canvas item (selection_box)
		self.group.selection_box.set_data('node', self)
		self.group.selection_box.set_data('terminal', None)

		self.group.selection_box.connect('event', self.network.node_event, self)

		self.group.selection_box.hide()
		
	def _add_initial_inputs(self):
		"""
		This function is called by buildnodeGUI, and creates the inputs for the node, 
		initializing them allto be connected to the network node (network.node0)
		"""


		for value in range(self.number_inputs_total):
			terminal = PiedTerminal(self)
			terminal.add_input()

	def _add_initial_outputs(self):
		"""
		This function is called by buildnodeGUI, and creates the outputs for the node, 
		initializing them all to be connected to the network node (network.node0)
		"""
		for value in range(self.number_outputs_total):
			terminal = PiedTerminal(self)
			terminal.add_output()

	def _arrangeGUI(self):
		"""
		This function is called by buildnodeGUI, and just arranges the added
		GUI components so they look pretty
		"""
		self.group.selection_box.lower_to_bottom()
		# lines ???
		self.group.text_box.raise_to_top()
		self.group.text.raise_to_top()
		self.group.dot.raise_to_top()
		self.group.dot_text.raise_to_top()
		
		self.group.symbol.raise_to_top()
		self.group.icon.raise_to_top()
		
	def _addInitialWidget(self):
		"""
		Called by buildnodeGUI, this adds the initial widget to the 
		node.
		"""
# JWB - windowlet code moved to separate module

		# import the right module and get the gui
		# 1/9/00==special case for container widget
		if (self.gui_name == 'container'):
			exec 'from widgets.' + self.gui_name + ' import *'
			self.gui = PiedWidget(self)
		# special case for prog_select Widget--needs to recieve the 
		# node
		elif (self.gui_name == 'prog_select'):
			exec 'from widgets.' + self.gui_name + ' import *'
			self.gui = PiedWidget(self)
		# special case for file_select Widget
		elif (self.gui_name == 'file_select'):
			exec 'from widgets.' + self.gui_name + ' import *'
			self.gui = PiedWidget(self)
		elif (self.gui_name == 'savecontainer'):
			exec 'from widgets.' + self.gui_name \
			  + ' import *'
			self.gui = PiedWidget(self)
			self.gui.show_contents()

		# special case for subnet nodes, unique location for the widget.
		# Also, need to pass the connection to the middle and the 
		# network name where the new nodes is being created.
# JWB - I want to make an actual network widget, so most of this code can
# be reduced
		elif (self.gui_name == 'network'):
			exec 'from Network import *'
			self.gui = PiedWidget(self.network.document, 
			  self.network, 'a_temp_id')
			
		elif self.gui_name == 'default':
			exec 'from modules.widgets.default import *'
			self.gui = PiedWidget(self)
		else:
			exec 'from modules.widgets.' + self.gui_name + ' import *'
			self.gui = PiedWidget()

# JWB - windowlet code moved to separate module

		windowlet_width = self.gui.width + 30 # margin
		windowlet_height = self.gui.height + 30 # margin
		windowlet_x = self.world_x
		windowlet_y = self.text_y + self.group.text['text_height'] / 2 + self.space
		self.windowlet = PiedWindowlet(windowlet_width,
										windowlet_height,
										windowlet_x,
										windowlet_y)
		self.windowlet.set_gui(self.gui)
		self.windowlet.set_message(self.message)
		self.windowlet.add_to_group(self.group)

		self.gui.show(self.network)

		# toggle that this node is selected
		self.network.toggle_selected(self)
		
		# connect gui events to an event handler to avoid competition
		# between the gui event handlers and the main event handlers
		# clicks		
		self.gui.set_data('node', self)
		self.gui.connect('event', self.guiEvent)
		
	def fixPosition(self, x, y):
		"""
		This function resets the position of the entire node at the
		specified position. This is useful because there can be a 
		discrepency between the world_x and the world_y and the
		actual position of the node on the canvas, because canvas
		coordinates are integers, and world coordinates are doubles.
		*****Note: this doesn't work right yet--so don't use it! Brad
		1/28/00
		"""
		# reset the world x and y
		self.world_x = x
		self.world_y = y

		# now reset the position of all the groups.
		self.dot_x = self.world_x
		self.dot_y = self.world_y
		self.group.dot_text.set(x = self.dot_x, y = self.dot_y)

		self.group.dot.set(x1=self.dot_x-self.dot_half_width,  
                   y1=self.dot_y-self.dot_half_width,  
                   x2=self.dot_x+self.dot_half_width,  
                   y2=self.dot_y+self.dot_half_width)

		self.group.symbol.set(x = self.world_x, y = self.world_y)
		self.group.icon.set(x = self.world_x, y = self.world_y)
		self.group.text.set(x = self.world_x, y = self.world_y)
		self.group.text_box.set(x1 = self.text_x - self.text_box_half_width,
			y1 = self.text_y - self.text_box_half_height,
			x2 = self.text_x + self.text_box_half_width,
			y2 = self.text_y + self.text_box_half_height)

		node_bounds = self.group.get_bounds()
		self.group.selection_box.set(x1 = node_bounds[0] - self.space,
			     y1 = node_bounds[1] - self.space,
			     x2 = node_bounds[2] + self.space,
			     y2 = node_bounds[3] + self.space)


		windowlet.ebox_x = self.world_x
		windowlet.ebox_y = self.text_y + self.group.text['text_height'] / 2 + self.gui.height / 2 + self.space
	
		self.group.window_widget.set(x = windowlet.ebox_x,
		      y = windowlet.ebox_y)

		
	def guiEvent(self, widget, event):
		"""
		This is just an event handler to avoid competition between
		the gui event handler and the main canvas event handlers. 
		This causes button 3 signals not to be propogated beyond
		the current gui.
		"""
		# if we have a popup menu press, we need to return true to 
		# prevent the signal from being passed on to underlying
		# networks.
		if ((event.type == GDK.BUTTON_PRESS)):
			if (event.button == 3):
				return TRUE
			
	def showGUI(self):
		"""
		This shows the current GUI for the node.
		"""
		self.group.window_widget.grab_focus()
		self.windowlet.ebox.show()
		self.group.window_widget.show()
		self.windowlet.shown = TRUE
		self.network.toggle_selected(self)

	def hideGUI(self):
		"""
		This hides the current GUI for the node.
		"""
		self.windowlet.ebox.hide()
		self.group.window_widget.hide()
		self.windowlet.shown = FALSE

	def set_gui(self, newGUI):
		"""
		Takes the passed Widget newGUI and sets this as the new node.gui
		element of the widget. The window is hidden by default.
		"""
		self.hideGUI()
		
		# resize and reset the window for the new widget
		self.group.window_widget.set(width = newGUI.width, 
		height = newGUI.height, 
		y = self.text_y + self.group.text['text_height'] / 2 + newGUI.height / 2 + self.space)
		
		# get rid of the old gui	
		self.windowlet.pane.remove(self.scrolled_window)
		
		# add the new gui
		self.gui = newGUI
		
		self.scrolled_window = GtkScrolledWindow()
		self.scrolled_window.set_border_width(5)
		self.scrolled_window.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
		self.scrolled_window.set_usize(self.network.width, self.network.height)
		self.windowlet.pane.add(self.scrolled_window)
		
		self.scrolled_window.add_with_viewport(self.gui)

		# attach the new gui to the event handlers which prevent button
		# 3 clicks from being propogated to the main network.
		self.gui.set_data('node', self)
		self.gui.connect('event', self.guiEvent)
		
		self.scrolled_window.show()
		self.gui.show(self.network)
		
		# toggle that this node is selected
		self.network.toggle_selected(self)
		
	def setText(self, textToSet):
		"""
		Sets the passed text as the text label of the node.
		"""
		self.group.text.set(text=textToSet)
		self.label = textToSet
		# reset the text box to fit the new label
			
	def setSelectColor(self, color):
		"""
		This can be used to set a user selected color for some elements (to 
		allow some customization of the GUI). The selected color is set
		as the passed argument color.
		"""
		self.color_selected = color
		
	def setIconTheme(self, theme):
		"""
		Change the theme of an icon to the passed theme arguement.
		"""
		self.icon_theme = theme
		
	def setIconName(self, name):
		"""
		Change the name of an icon to the passed name argument.
		"""
		self.icon_name = name
		
	def setSymbolTheme(self, theme):
		"""
		Change the theme of a symbol to the passed theme argument.
		"""
		self.symbol_theme = theme
										   
	def setColors(self, bg, fg):
		""" 
		Set the node background and foreground colors.
		"""
		self.color_bg = bg
		self.color_fg = fg



