"""savecontainer.py

A widget for implementing a permanent storage container.

Classes:
	LociWidget
"""
# standard modules
import os

# pyGtk/pyGnome
from gtk import *
from gnome.ui import *
from GDK import *

#the 4DOM DOM implementation
from xml.dom import DOMImplementation
from xml.dom import ext

# locus modules
from modules.handlemiddleresponse import *
from modules.xml2gui import Xml2Gui

# configuration variables
from piperconfig import *

# constants
SMALL_PIXMAP_DIR = os.path.join(piper_pixmap_dir, 'file')

class PiedWidget(GtkVBox):
	"""A container class displaying permanently stored loci.
	
	This class inherets from GtkVBox, but should eventually inheret
	from a base container class.
	
	The info in the container is displayed using a GtkTree.
	"""
	get_type = GtkVBox(spacing=5).get_type()
	
	def __init__(self, locus):
		"""Initialize a basic GUI representing the save directory of Loci.
		
		Arguments:
			locus - The locus that this GUI represents.
		"""
		self._o = GtkVBox(spacing=5)._o
		
		self.locus = locus
		
		self.width = 260
		self.height = 260
		
		# the directory that the savecontainer represents (relative to
		# SAVE_DIR)
		self.dir_represented = ''
		
		# a frame to hold the info about the name of the directory
		self.frame = GtkFrame()
		self.frame.show()
		
		# the GtkCTree which will display the info
		self.save_tree = GtkCTree()
		from container_tree import *
		self.tree_view = LociWidget(self.save_tree)
		self.add(self.frame)
		self.frame.add(self.tree_view)
		
		# connect the tree for drags and drops
		# sending
		self.save_tree.connect("select_row", self.begin_drag)
		
		# receiving
		workspace_locus = [('WSLOCUS', 0, 0)]
		self.save_tree.drag_dest_set(DEST_DEFAULT_ALL, workspace_locus, 
		  ACTION_MOVE)
		self.save_tree.connect('drag_data_received', self.add_dropped_locus)
		
	def set_contents(self, directory):
		"""Set the contents of the widget to display a particular directory.
		
		Arguments:
			directory - The directory to display in the saved Loci info.
		"""
		self.dir_represented = directory
	
	def show_contents(self):
		"""Shows the contents of the currently set directory.
		
		Calls the middle to retrieve the directory contents of the directory
		and displays these contents.
		
		Raises an IOError if the current dir_represented does not exist.
		"""
		# create the info request
		di = DOMImplementation.DOMImplementation()
		dom_tree = di.createDocument(None, None, None)
			
		front_node = dom_tree.createElement('front')
		dom_tree.appendChild(front_node)
		
		info_node = dom_tree.createElement('get_info')
		front_node.appendChild(info_node)
		
		locus_node = dom_tree.createElement('locus')
		locus_node.setAttribute('id', self.locus.name)
		info_node.appendChild(locus_node)
		
		# send the xml to the middle
		self.locus.workspace.middle_connect.send_dom(dom_tree)
		
		ext.ReleaseNode(dom_tree)
		
		# get the reply
		reply = self.locus.workspace.middle_connect.wait_for_reply()
		reply_handler = GetInfoHandler(reply)
		
		if (reply_handler.is_success() == 1):
			dom_info = reply_handler.get_info()
			gui_builder = Xml2Gui(dom_info)
			save_items = gui_builder.get_info()
			for n in save_items:
				self._add_item_to_tree(n[0], n[1], n[2])
		
		# otherwise pass the error on to the calling function
		else:
			error_num = reply_handler.is_error()
			if (error_num == '6'):
				raise IOError(self.dir_represented + ' was not found.')
				
	def _add_item_to_tree(self, name_to_add, pix, mask, item_to_add_to = None):
		"""Adds a new item into the tree representing items in the container.
		
		Arguments:
			name_to_add - The name of the item to add.
			pix - The picture next to the item
			mask - The annoying thing that always has to be with the pix.
			level_to_add - The item to add the item to in the tree. If this
			is none, the item will be added at the base of the tree.
		"""
		# add the element to the list
		self.save_tree.insert_node(item_to_add_to, None, [name_to_add,], 2, 
		  pix, mask, pix, mask, FALSE, TRUE)
							  
	def begin_drag(self, widget, row, column, z=None):
		"""Connect row selection with possible drag and drop.
		"""
		print 'dnd'
		
	def add_dropped_locus(self, w, context, x, y, data, info, time):
		"""Recieve a drop onto the tree and add the new item.
		
		Needs to recieve a string with the name of the locus to be added to
		the save directory (ie. locus.name), with the name of the locus on
		the screen and with the type of the locus. These are assumed to be 
		passed as three items separated by '*'s: 
			'locus_name*screen_name*type'
		
		The middle is then called to add this locus to the save directory.
		"""
		if data and data.format == 8:
			# get the name and screen_name
			locus_name, screen_name, type = string.split(data.data, '*')
			# finish off the drop
			w.drag_finish(context, TRUE, FALSE, time)
			
			# case 1--> containers (since they are directories and can contain
			# other loci)
			if (type == 'container'):
				pass
			# case 2 -> composite (since they have loci in their workspace)
			elif (type == 'composite'):
				pass
			# case 3--> we have a normal type (not composite or container)
			else:
				# call the middle to add the item to the save directory
				di = DOMImplementation.DOMImplementation()
				dom_tree = di.createDocument(None, None, None)
			
				front_node = dom_tree.createElement('front')
				dom_tree.appendChild(front_node)
			
				save_node = dom_tree.createElement('save')
				front_node.appendChild(save_node)
			
				con_node = dom_tree.createElement('container')
				con_node.setAttribute('id', self.locus.name)
				save_node.appendChild(con_node)
			
				locus_node = dom_tree.createElement('locus')
				locus_node.setAttribute('id', locus_name)
				locus_node.setAttribute('name', screen_name)
				save_node.appendChild(locus_node)
				
			# now send the XML to the middle
			self.locus.workspace.middle_connect.send_dom(dom_tree)
			
			ext.ReleaseNode(dom_tree)
			
			reply = self.locus.workspace.middle_connect.wait_for_reply()
			reply_handler = BaseHandler(reply)
			# if the addition was successful, add the new locus to the tree
			if (reply_handler.is_success() == 1):
				get_gui = Xml2Gui()
				pix, mask = get_gui.get_pix_mask(type)
				self._add_item_to_tree(screen_name, pix, mask)
			else:
				raise IOError('Problem with adding the locus.')
		
		
		
