###########################################################################
# Many thanks to the xml-sig page for providing lots of good code to use!
###########################################################################

# standard modules
import os

# pyGtk
from gtk import *

#the 4DOM DOM implementation
from xml.dom import ext
from xml.dom.NodeFilter import NodeFilter
from xml.dom.ext.reader import Sax
from xml.dom import DOMImplementation

# configuration variables
from piperconfig import *

####constants
SMALL_PIXMAP_DIR = os.path.join(piper_pixmap_dir, 'file')

class Xml2Gui:
	"""General class for converting xml for container info into a gui.
	
	This class takes as input a dom_tree (passed from the dl) describing
	xml of the following format:
		<container_info type = 'container_type' name = 'name of container'>
		  <element1 name = 'name' />
		  <element2 name = 'name' />
		  ....
		</container_info>
	
	It will then parse this and generate a tree representation of the 
	contents for display in the container.
	"""
	
	def __init__(self, dom_info = None):
		"""Setup up the converter with a dom tree for parsing.
		
		Arguments:
			dom_info - dom info in the format described above.
			
		Sets up the pixmaps which will be used for filling in the
		tree icons.
		"""
		if dom_info:
			di = DOMImplementation.DOMImplementation()
			self.dom_tree = di.createDocument(None, None, None)
		
			# import the dom info into our tree
			info_node = self.dom_tree.importNode(dom_info, 1)
			self.dom_tree.appendChild(info_node)

		self._setup_pixmaps()
		
	def make_tree(self):
		"""Return the dom info as a GtkCTree.
		
		Returns:
			A GtkCTree containing the elements of the container.
		"""
		# create our tree
		self.info_tree = GtkCTree()
		
		dom_iterator = self.dom_tree.createNodeIterator(self.dom_tree,
		  NodeFilter.SHOW_ELEMENT, None, 0)
		  
		cur_node = dom_iterator.nextNode()
		
		# iterate through the dom and set the appropriate elements
		while cur_node:
			# if we are at the top level, add the name of the container
			# as the column title
			if (cur_node.nodeName == 'container_info'):
				self.info_tree.set_column_title(0, 
				  cur_node.getAttribute('name'))
			# otherwise add the item to the list
			else:
				self._add_item_to_tree(cur_node.getAttribute('name'),
				  cur_node.nodeName)
				  
			cur_node = dom_iterator.nextNode()
			
		return self.info_tree
		
	def get_info(self):
		"""Return the information in the dom tree in a gui format.
		
		Returns:
			A 2d list. Each element of the list has the following three
			components:
				o item name
				o item pix
				o item mask
		
		This function can be used for getting pixmap and name information
		to do a "customized" tree or list insertion.
		"""
		total_info = []
		dom_iterator = self.dom_tree.createNodeIterator(self.dom_tree,
		  NodeFilter.SHOW_ELEMENT, None, 0)
		  
		cur_node = dom_iterator.nextNode()
		
		# iterate through the dom and set the appropriate elements
		while cur_node:
			if (cur_node.nodeName == 'container_info'):
				pass
			# otherwise add the item to the list
			else:
				inner_list = []
				inner_list.append(cur_node.getAttribute('name'))
				pix, mask = self._get_pixmap(cur_node.nodeName)
				inner_list.append(pix)
				inner_list.append(mask)
				total_info.append(inner_list)
				
			cur_node = dom_iterator.nextNode()
			
		return total_info
		
	def get_pix_mask(self, type):
		"""Return the pix and mask for passed type."""
		return self._get_pixmap(type)
		
	# internal functions
	def _setup_pixmaps(self):
		"""Create the little pixmaps for displaying with container items.
		"""
		pix_win = GtkWindow()
		# directories
		self.dir_pix, self.dir_mask = create_pixmap_from_xpm(pix_win, None, 
		  os.path.join(SMALL_PIXMAP_DIR, "bookclosed.xpm")) 
		self.odir_pix, self.odir_mask = create_pixmap_from_xpm(pix_win, 
		  None, os.path.join(SMALL_PIXMAP_DIR, "bookopen.xpm"))
		  
		# files
		self.file_pix, self.file_mask = create_pixmap_from_xpm(pix_win, 
		  None, os.path.join(SMALL_PIXMAP_DIR, "minipage.xpm"))
		
		# databases
		self.db_pix, self.db_mask = create_pixmap_from_xpm(pix_win, 
		  None, os.path.join(SMALL_PIXMAP_DIR, "smalldb.xpm"))
		self.table_pix, self.table_mask = create_pixmap_from_xpm(pix_win, 
		  None, os.path.join(SMALL_PIXMAP_DIR, "folder.xpm")) 
		  
		# default
		self.default_pix, self.default_mask = create_pixmap_from_xpm(pix_win,
		  None, os.path.join(SMALL_PIXMAP_DIR, "default.xpm"))
		  
	def _add_item_to_tree(self, name_to_add, type_to_add, item_to_add_to = None):
		"""Adds a new item into the tree representing items in the container.
		
		Arguments:
			name_to_add - The name of the item to add.
			type_to_add - The type of the item to add.
			level_to_add - The item to add the item to in the tree. If this
			is none, the item will be added at the base of the tree.
		"""
		# first get the appropriate pixmap to display with the GUI
		pix, mask = self._get_pixmap(type_to_add)
			
		# add the element to the list
		self.info_tree.insert_node(item_to_add_to, None, [name_to_add,], 2, 
		  pix, mask, pix, mask, FALSE, TRUE)
		  
	def _get_pixmap(self, locus_type):
		"""Return an appropriate pixmap corresponding to a passed type.
		
		Arguments:
			locus_type - A string corresponding to one of the different
			defined locus types.
			
		Each locus_type is associated with a particular little pixmap so
		that they can be distinguished in the GUI.
		"""
		if (locus_type == 'directory'):
			return self.dir_pix, self.dir_mask
		elif (locus_type == 'file') or (locus_type == 'document'):
			return self.file_pix, self.file_mask
		elif (locus_type == 'table'):
			return self.table_pix, self.table_mask
		elif (locus_type == 'database'):
			return self.db_pix, self.db_mask
		else:
			return self.default_pix, self.default_mask

